from pnb.mcl.test.model_api import AbstractTestCase


class Test_Enumeration(AbstractTestCase):

    def test_sortings_when_sorted(self):
        e = self.Enumeration('e')
        ea = self.EnumerationLiteral('a', e)
        eb = self.EnumerationLiteral('b', e)
        self.assertEqual(
            list(e.ownedLiterals),
            [ea, eb])
        self.assertEqual(
            list(e.orderedOwnedLiterals),
            [ea, eb])

    def test_sortings_when_not_sorted(self):
        e = self.Enumeration('e')
        eb = self.EnumerationLiteral('b', e)
        ea = self.EnumerationLiteral('a', e)
        self.assertEqual(
            list(e.ownedLiterals),
            [ea, eb])
        self.assertEqual(
            list(e.orderedOwnedLiterals),
            [eb, ea])



from unittest import TestCase
from lxml import etree
from pnb.mcl.metamodel import standard
from pnb.mcl.io.xml import XmlExporter, XmlImporter
from pnb.mcl.test.examples import ex_1


class Test_XmlImporter_ex_1(TestCase):

    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        cls.model_by_uri = ex_1.make_model_by_uri(standard)
        cls.xml_by_uri = {uri: XmlExporter(model).xml for uri, model in cls.model_by_uri.items()}
        cls.xml_loader = staticmethod(lambda uri: cls.xml_by_uri[uri])

    def _test_importer(self, uri):
        messages = []
        importer = XmlImporter(self.xml_loader, uri)
        for model in importer.model_by_uri.values():
            reexported_xml_code = etree.tostring(
                XmlExporter(model).xml, encoding='unicode', pretty_print=True)
            expected_xml_code = etree.tostring(
                self.xml_by_uri[model.uri], encoding='unicode', pretty_print=True)
            if reexported_xml_code != expected_xml_code:
                messages += [
                    f'-------- {model.uri} --------\n\n--- reexported ---\n',
                    reexported_xml_code,
                    '--- expected ---\n',
                    expected_xml_code]
        if messages:
            print(f'\n\n########## {self} ##########\n')
            print('\n'.join(messages))
            self.fail()

    def test_ModelCore(self):
        self._test_importer('http://ModelCore')

    def test_ModelProcess(self):
        self._test_importer('http://ModelProcess')

    def test_ModelInstance1(self):
        self._test_importer('http://ModelInstance1')

    def test_ModelInstance2a(self):
        self._test_importer('http://ModelInstance2a')

    def test_ModelInstance2b(self):
        self._test_importer('http://ModelInstance2b')

    def test_ModelInstance3(self):
        self._test_importer('http://ModelInstance3')



import abc
import inspect
import pathlib
from unittest import SkipTest, TestCase, TestSuite

THIS_DIR = pathlib.Path(__file__).parent

class AbstractTestCase(TestCase):

    _is_abstract = True

    @classmethod
    def setUpClass(cls):
        if cls._is_abstract:
            raise SkipTest(f'{cls} is an abstract base class')
        super().setUpClass()
        model = cls.get_model()
        cls.CompositionProperty = model.CompositionProperty
        cls.ConcreteClass = model.ConcreteClass
        cls.Enumeration = model.Enumeration
        cls.EnumerationLiteral = model.EnumerationLiteral
        cls.Members = model.Members
        cls.MutableMembers = model.MutableMembers
        cls.Namespace = model.Namespace
        cls.ReferenceProperty = model.ReferenceProperty
        cls.Package = model.Package

    @classmethod
    @abc.abstractmethod
    def get_model(cls):
        ...


def find_abstract_test_classes(suite):
    return [cls for cls in set(iter_test_classes(suite))
            if issubclass(cls, AbstractTestCase) and cls._is_abstract]

def iter_test_classes(suite):
    for test in suite._tests:
        if isinstance(test, TestSuite):
            yield from iter_test_classes(test)
        elif isinstance(test, TestCase):
            yield type(test)
        else:
            raise TypeError(test)


def get_suite(loader, model):
    # TODO
    # - better way to find abstract test classes
    #   (loader.discover creates instances!) -> check private methods of loader
    # - pattern from call???
    api_suite = loader.discover(str(THIS_DIR), pattern='*.py')
    suite = TestSuite()
    for cls in find_abstract_test_classes(api_suite):
        ConreteTest = type(
            f'{cls.__name__}_{model.__name__.rsplit(".", 1)[-1]}',
            (cls, ),
            dict(_is_abstract=False, get_model=classmethod(lambda cls: model)))
        suite.addTests(loader.loadTestsFromTestCase(ConreteTest))
    return suite


def load_suite(model):
    def load_tests(loader, tests, pattern):
        return get_suite(loader, model)
    return load_tests



import itertools
import sys

import rdflib
from rdflib import Literal, RDF, BNode


class RdfExporter:

    def __init__(self, model, metamodel_namespace = 'http://www.plants-and-bytes.de/mcl/meta/'):
        
        self.model = model
        
        self.meta_model = sys.modules[model.__module__]
        self.graph = rdflib.Graph()
        self.node_by_element = {}
        self.meta_model_namespace = rdflib.Namespace(metamodel_namespace)
        self.graph.bind('meta', self.meta_model_namespace)
        self.add_element(model)

        
    def get_element_node(self, element):
        
        assert isinstance(element, self.meta_model.Element), element
        
        node = self.node_by_element.get(element)
        if node is None:
            if isinstance(element, self.meta_model.Model):
                if element.uri is None:
                    assert element.name == 'builtin'
                    uri = 'http://www.plants-and-bytes.de/mcl/builtin'
                else:
                    uri = element.uri
                node = rdflib.URIRef(uri)
                self.graph.bind(element.name, node+'/')
            elif isinstance(element, self.meta_model.Object):
                name = element.name
                if name:
                    node = rdflib.URIRef(self.get_element_node(element.model) + '/' + name) # TODO: check unique
                else:
                    model = element.model
                    assert model is self.model
                    node = BNode()
            else:
                owner = element.owner
                owner_node = self.get_element_node(owner)
                if isinstance(owner, self.meta_model.Model):
                    intro = owner_node + '/'
                else:
                    intro = owner_node + '.'
                node = rdflib.URIRef(intro + element.name)

            
            self.node_by_element[element] = node
            
        return node

    def add(self, s, p, o):
        triple = s, p, o
        if None not in triple:
            self.graph.add(triple)
            
            
    def extend(self, node, values, property_):
        
        for value in values:
            assert isinstance(value, rdflib.Node)
            self.add(node, property_, value)
            
            
    def extend_with_list(self, node, elements, property_):
        self.extend(node, elements, property_) # TODO
        
        
    def add_element(self, element):
        
        assert isinstance(element, self.meta_model.Element), element

        node = self.get_element_node(element)
        self.add(node, RDF.type, self.meta_model_namespace.term(element.get_meta_class_name()))

        if isinstance(element, self.meta_model.NamedElement):
            if element.name:
                self.add(node, self.meta_model_namespace.name, Literal(element.name))

        if isinstance(element, self.meta_model.Package):
            self.extend(node, [self.add_element(pe) for pe in element.packagedElements], self.meta_model_namespace.packagedElement)
            
        if isinstance(element, self.meta_model.Model):
            self.extend(node, [self.add_element(pe) for pe in element.packagedElements], self.meta_model_namespace.packagedElement)
            self.extend(node, [self.add_element(pe) for pe in element.unnamedObjects], self.meta_model_namespace.packagedElement)

        if isinstance(element, self.meta_model.Type):
            self.extend(node, [self.get_element_node(st) for st in element.superTypes], self.meta_model_namespace.superType)
            self.extend(node, [self.add_element(at) for at in element.ownedAttributes], self.meta_model_namespace.ownedAttribute)

        if isinstance(element, self.meta_model.Property):
            self.add(node, self.meta_model_namespace.type, self.get_element_node(element.type))
            self.add(node, self.meta_model_namespace.term('lower'), Literal(element.lower))
            if element.upper is not None:
                self.add(node, self.meta_model_namespace.term('upper'), Literal(element.upper))
            self.add(node, self.meta_model_namespace.isOrdered, Literal(element.isOrdered))
            if isinstance(
                    element, (self.meta_model.ReferenceProperty, self.meta_model.DataProperty)):
                self.add(node, self.meta_model_namespace.isUnique, Literal(element.isUnique))
            if isinstance(element, self.meta_model.ReferenceProperty):
                self.add(node, self.meta_model_namespace.term('oppositeLower'),
                    Literal(element.oppositeLower))
                if element.oppositeUpper is not None:
                    self.add(node, self.meta_model_namespace.term('oppositeUpper'),
                        Literal(element.oppositeUpper))

        if isinstance(element, self.meta_model.Enumeration):
            if 0:
             self.extend(node, [self.add_element(lit) for lit in element.packagedElements], self.meta_model_namespace.packagedElement)
             raise Exception('enums not yet supported')
             # TODO
             add_children(item.orderedOwnedLiterals)


        if isinstance(element, self.meta_model.Object):
            self.add(node, self.meta_model_namespace.type, self.get_element_node(element.type))
            
            #attributes['type'] = self.get_reference(item.type)
            for prop in element.type.attributes.values():
                values = prop._get_values_(element)
                if values:
                    if prop.isUnique and not prop.isOrdered:
                        add_values = self.extend
                    else:
                        # better choice in some cases?
                        add_values = self.extend_with_list
                    if isinstance(prop, self.meta_model.DataProperty):
                        handle_value = lambda value: self.add_data_value(prop.type, value)
                    elif isinstance(prop, self.meta_model.CompositionProperty):
                        handle_value = self.add_element # TODO: check model
                    else:
                        handle_value = self.get_element_node

                    add_values(node, [handle_value(value) for value in values], self.get_element_node(prop))

        return node

    def add_data_value(self, data_type, value):
        if isinstance(value, (bool, str, int, float)):
            return rdflib.Literal(value)

        if isinstance(value, self.meta_model.EnumerationLiteral):
            return self.get_reference(value)
        else:
            raise TypeError(f'{type(value)} is not yet supported')

    def get_reference(self, element):
        model, qname = element.get_model_and_qname()
        if qname is None:
            if model is not self.model:
                model, qname = element.get_model_and_qname()
                raise Exception('TODO')
            id_ = self._id_by_element.get(element)
            if id_ is None:
                try:
                    type_name = element.type.name
                except AttributeError:
                    raise # TODO
                nr = self._id_count_by_type_name.get(type_name, 0) + 1
                self._id_count_by_type_name[type_name] = nr
                id_ = f'#{type_name}{nr}'
                self._id_by_element[element] = id_
            return id_
        else:
            return rdflib.Literal(model.uri + '/' + qname)


    def get_model_reference(self, model):
        reference = self.reference_by_model.get(model)
        if reference is None:
            reference = self._find_free_prefix(model.name)
            self.reference_by_model[model] = reference
        return reference

    def _find_free_prefix(self, prefix):

        def suffixes():
            yield ''
            for nr in itertools.count():
                yield str(nr)
        for suffix in suffixes():
            candidate = prefix + suffix
            if candidate not in self.reference_by_model.values():
                return candidate




from unittest import TestCase

from rdflib import Graph
from rdflib.compare import graph_diff, to_isomorphic

from pnb.mcl.metamodel import standard
from pnb.mcl.io.rdf import RdfExporter
from pnb.mcl.test.examples import ex_1

def assertGraphsEqual(test, graph_1, turtle_2):
    iso_1 = to_isomorphic(graph_1)
    iso_2 = to_isomorphic(Graph().parse(format='turtle', data=turtle_2))
    _, in_iso_1, in_iso_2 = graph_diff(iso_1, iso_2)
    message_parts = []
    if in_iso_1:
        message_parts.append(' *** only in 1 *** ')
        message_parts.extend(repr(triple) for triple in in_iso_1)
    if in_iso_2:
        message_parts.append(' *** only in 2 *** ')
        message_parts.extend(repr(triple) for triple in in_iso_2)
    if message_parts:
        test.fail('\n'.join(message_parts))


class Test_XmlExporter_ex_1(TestCase):

    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        cls.model_by_uri = ex_1.make_model_by_uri(standard)

    def test_ModelCore(self):
        exporter = RdfExporter(self.model_by_uri['http://ModelCore'])
        assertGraphsEqual(self, exporter.graph, '''
            @prefix ModelCore: <http://ModelCore/> .
            @prefix meta: <http://www.plants-and-bytes.de/mcl/meta/> .

            <http://ModelCore> a meta:Model ;
                meta:name "ModelCore" ;
                meta:packagedElement ModelCore:ConceptualModel .

            ModelCore:ConceptualModel a meta:AbstractClass ;
                meta:name "ConceptualModel" .''')

    def test_ModelProcess(self):
        exporter = RdfExporter(self.model_by_uri['http://ModelProcess'])
        assertGraphsEqual(self, exporter.graph, '''
            @prefix ModelCore: <http://ModelCore/> .
            @prefix ModelProcess: <http://ModelProcess/> .
            @prefix builtin: <http://www.plants-and-bytes.de/mcl/builtin/> .
            @prefix meta: <http://www.plants-and-bytes.de/mcl/meta/> .

            <http://ModelProcess> a meta:Model ;
                meta:name "ModelProcess" ;
                meta:packagedElement ModelProcess:Base,
                    ModelProcess:Steps .

            ModelProcess:Base a meta:Package ;
                meta:name "Base" ;
                meta:packagedElement ModelProcess:Base.ProcessStep .

            ModelProcess:Base.ProcessStep a meta:ConcreteClass ;
                meta:name "ProcessStep" ;
                meta:ownedAttribute ModelProcess:Base.ProcessStep.RelatedSteps,
                    ModelProcess:Base.ProcessStep.StepNumber,
                    ModelProcess:Base.ProcessStep.SubSteps ;
                meta:superType ModelCore:ConceptualModel .

            ModelProcess:Base.ProcessStep.RelatedSteps a meta:ReferenceProperty ;
                meta:name "RelatedSteps" ;
                meta:type ModelProcess:Base.ProcessStep ;
                meta:lower 0 ;
                meta:isOrdered false ;
                meta:isUnique true ;
                meta:oppositeLower 0 .

            ModelProcess:Base.ProcessStep.StepNumber a meta:DataProperty ;
                meta:name "StepNumber" ;
                meta:type builtin:String ;
                meta:lower 0 ;
                meta:upper 1 ;
                meta:isOrdered false ;
                meta:isUnique false .         

            ModelProcess:Base.ProcessStep.SubSteps a meta:CompositionProperty ;
                meta:name "SubSteps" ;
                meta:type ModelProcess:Base.ProcessStep ;
                meta:lower 0 ;
                meta:isOrdered false .

            ModelProcess:Steps a meta:Package ;
                meta:name "Steps" ;
                meta:packagedElement ModelProcess:Steps.Pumping .

            ModelProcess:Steps.Pumping a meta:ConcreteClass ;
                meta:name "Pumping" ;
                meta:superType ModelProcess:Base.ProcessStep .''')

    def test_ModelInstance1(self):
        exporter = RdfExporter(self.model_by_uri['http://ModelInstance1'])
        assertGraphsEqual(self, exporter.graph, '''
            @prefix ModelInstance2: <http://ModelInstance2a/> .
            @prefix ModelInstance21: <http://ModelInstance2b/> .
            @prefix ModelProcess: <http://ModelProcess/> .
            @prefix meta: <http://www.plants-and-bytes.de/mcl/meta/> .

            <http://ModelInstance1> a meta:Model ;
                meta:name "ModelInstance1" ;
                meta:packagedElement [ a meta:Object ;
                        ModelProcess:Base.ProcessStep.RelatedSteps ModelInstance2:Step1,
                            ModelInstance21:Step ;
                        ModelProcess:Base.ProcessStep.StepNumber "step1" ;
                        meta:type ModelProcess:Steps.Pumping ] .''')

    def test_ModelInstance2a(self):
        exporter = RdfExporter(self.model_by_uri['http://ModelInstance2a'])
        assertGraphsEqual(self, exporter.graph, '''
            @prefix ModelInstance2: <http://ModelInstance2a/> .
            @prefix ModelInstance3: <http://ModelInstance3/> .
            @prefix ModelProcess: <http://ModelProcess/> .
            @prefix meta: <http://www.plants-and-bytes.de/mcl/meta/> .
            
            <http://ModelInstance2a> a meta:Model ;
                meta:name "ModelInstance2" ;
                meta:packagedElement _:Pumping1,
                    ModelInstance2:Step1 .
            
            ModelInstance2:Step1 a meta:Object ;
                ModelProcess:Base.ProcessStep.RelatedSteps _:Pumping1,
                    ModelInstance3:Step ;
                ModelProcess:Base.ProcessStep.StepNumber "step2a1" ;
                meta:name "Step1" ;
                meta:type ModelProcess:Steps.Pumping .
            
            _:Pumping1 a meta:Object ;
                ModelProcess:Base.ProcessStep.StepNumber "step2a2" ;
                meta:type ModelProcess:Steps.Pumping .''')

    def test_ModelInstance2b(self):
        exporter = RdfExporter(self.model_by_uri['http://ModelInstance2b'])
        assertGraphsEqual(self, exporter.graph, '''
            @prefix ModelInstance2: <http://ModelInstance2b/> .
            @prefix ModelInstance21: <http://ModelInstance2a/> .
            @prefix ModelProcess: <http://ModelProcess/> .
            @prefix meta: <http://www.plants-and-bytes.de/mcl/meta/> .

            <http://ModelInstance2b> a meta:Model ;
                meta:name "ModelInstance2" ;
                meta:packagedElement ModelInstance2:Step .

            ModelInstance2:Step a meta:Object ;
                ModelProcess:Base.ProcessStep.RelatedSteps ModelInstance21:Step1 ;
                ModelProcess:Base.ProcessStep.StepNumber "step2b" ;
                meta:name "Step" ;
                meta:type ModelProcess:Steps.Pumping .''')

    def test_ModelInstance3(self):
        exporter = RdfExporter(self.model_by_uri['http://ModelInstance3'])
        assertGraphsEqual(self, exporter.graph, '''
            @prefix ModelInstance3: <http://ModelInstance3/> .
            @prefix ModelProcess: <http://ModelProcess/> .
            @prefix meta: <http://www.plants-and-bytes.de/mcl/meta/> .
            
            <http://ModelInstance3> a meta:Model ;
                meta:name "ModelInstance3" ;
                meta:packagedElement [ a meta:Object ;
                        ModelProcess:Base.ProcessStep.StepNumber "step31" ;
                        ModelProcess:Base.ProcessStep.SubSteps ModelInstance3:Step ;
                        meta:type ModelProcess:Steps.Pumping ] .
            
            ModelInstance3:Step a meta:Object ;
                ModelProcess:Base.ProcessStep.StepNumber "step32" ;
                meta:name "Step" ;
                meta:type ModelProcess:Steps.Pumping .''')



from collections import namedtuple
import itertools
import sys

from pnb.mcl.io.xmi import read_xmi
from lxml import etree

from datetime import datetime


class XmlExporter:

    def __init__(self, model):
        self.metamodel = sys.modules[model.__module__]
        self.model = model
        self.reference_by_model = {model: ''}
        
        self._id_by_element = {}
        self._id_count_by_type_name = {}
        self.xml_by_item = {}
        self.xml = self.item_to_xml(model)

        for element, id_ in self._id_by_element.items():
            xml = self.xml_by_item[element]
            assert 'id' not in xml.attrib
            assert 'name' not in xml.attrib
            attributes = {'id': id_[1:]}
            attributes.update(xml.attrib)
            xml.attrib.clear()
            xml.attrib.update(attributes)

        for ref_model, prefix in self.reference_by_model.items():
            if ref_model in (model, self.metamodel.BUILTIN):
                continue
            self.xml.insert(0, etree.Element(
                'Import',
                source=ref_model.uri,
                prefix=prefix))


    def item_to_xml(self, item):
        attributes = {}
        children = []


        def add_children(members):
            children.extend(self.item_to_xml(member) for member in members)

        if isinstance(item, self.metamodel.NamedElement):
            attributes['name'] = item.name
        if isinstance(item, self.metamodel.Model):
            attributes['uri'] = item.uri
            add_children(item.packagedElements)
            add_children(item.unnamedObjects)
        if isinstance(item, self.metamodel.Package):
            add_children(item.packagedElements)
        if isinstance(item, self.metamodel.Type):
            if item.superTypes:
                st_string = ' '.join(self.get_reference(type_) for type_ in item.superTypes)
                attributes['superTypes'] = st_string
            add_children(item.ownedAttributes)
        if isinstance(item, self.metamodel.Property):
            attributes['type'] = self.get_reference(item.type)
            attributes['lower'] = str(item.lower)
            attributes['upper'] = '*' if item.upper is None else str(item.upper)
            attributes['isOrdered'] = 'true' if item.isOrdered else 'false'
            if not item.get_meta_class_name() == 'CompositionProperty':
                attributes['isUnique'] = 'true' if item.isUnique else 'false'
            if item.get_meta_class_name() == 'ReferenceProperty':
                attributes['oppositeLower'] = str(item.oppositeLower)
                if item.oppositeUpper is None:
                    attributes['oppositeUpper'] = '*'
                else:
                    attributes['oppositeUpper'] = str(item.oppositeUpper)
        if isinstance(item, self.metamodel.Enumeration):
            add_children(item.orderedOwnedLiterals)
            
            
        if isinstance(item, self.metamodel.SingletonValue):
            attributes['type'] = self.get_reference(item.type)


        if isinstance(item, self.metamodel.Object):
            attributes['type'] = self.get_reference(item.type)
            for prop in item.type.attributes.values():
                values = prop._get_values_(item)
                if values:
                    match prop.get_meta_class_name():
                        case 'CompositionProperty':
                            child = etree.Element('Components')
                            child.extend(self.item_to_xml(value) for value in values)
                        case 'ReferenceProperty':
                            child = etree.Element('References')
                            child.attrib['refs'] = ' '.join(
                                self.get_reference(value) for value in values)
                        case 'DataProperty':
                            child = etree.Element('Data')
                            child.extend(self.data_to_xml(prop.type, value) for value in values)
                        case _:
                            continue
                        
                    child.attrib['property'] = prop.name
                    children.append(child)
                    
                    
        sorted_attributes = {}
        if name:= attributes.pop('name', None):
            sorted_attributes['name'] = name
        if id_:= attributes.pop('id', None):
            sorted_attributes['id'] = id_
        sorted_attributes.update(sorted(attributes.items()))

        xml = etree.Element(
            item.get_meta_class_name(),
            sorted_attributes)
        xml.extend(children)
        
        self.xml_by_item[item] = xml
        return xml
    
    
    def data_to_xml(self, data_type, value):
        attrs = {}
        if isinstance(value, str):
            tag = 'String'
            text = value
        elif isinstance(value, self.metamodel.EnumerationLiteral):
            tag = 'EnumerationLiteral'
            text = value.name
            attrs = {'type': self.get_reference(value.type)}
        elif isinstance(value, bool):
            tag = 'Boolean'
            text = str(value).lower()
        elif isinstance(value, int):
            # Keep this after bool!!!
            tag = 'Integer'
            text = str(value)
        elif isinstance(value, float):
            tag = 'Double'
            text = str(value)
        else:
            raise TypeError(value)
        
        xml = etree.Element(tag, attrs)
        xml.text = text
        return xml

    def get_reference(self, element):
        model, qname = element.get_model_and_qname()
        if qname is None:
            if model is not self.model:
                model, qname = element.get_model_and_qname()
                raise Exception('TODO')
            id_ = self._id_by_element.get(element)
            if id_ is None:
                try:
                    type_name = element.type.name
                except AttributeError:
                    raise # TODO
                nr = self._id_count_by_type_name.get(type_name, 0) + 1
                self._id_count_by_type_name[type_name] = nr
                id_ = f'#{type_name}{nr}'
                self._id_by_element[element] = id_
            return id_
                
            
        else:
            return f'{self.get_model_reference(model)}/{qname}'

    def get_model_reference(self, model):
        reference = self.reference_by_model.get(model)
        if reference is None:
            reference = self._find_free_prefix(model.name)
            self.reference_by_model[model] = reference
        return reference

    def _find_free_prefix(self, prefix):

        def suffixes():
            yield ''
            for nr in itertools.count():
                yield str(nr)
        for suffix in suffixes():
            candidate = prefix + suffix
            if candidate not in self.reference_by_model.values():
                return candidate


ModelInfo = namedtuple('ModelInfo', ['uri', 'xml_element', 'import_uris_by_prefix'])
ObjectInfo = namedtuple('ObjectInfo', ['built_object', 'xml_element'])
TypeInfo = namedtuple('TypeInfo', ['global_id', 'global_super_type_ids', 'xml_element'])
# separator symbols
ID_SEP_SYMBOL = '#'
SUB_ELEMENT_SEP_SYMBOL = '.'
URI_SEP_SYMBOL = '/'
# all types except PrimitiveTypes which are builtin mcl types
ALL_TYPES = ['AbstractClass', 'ConcreteClass',
             'AbstractDataType', 'AggregatedDataType', 'SingletonType',
             'Enumeration']

def sorted_by_dependency(type_infos):
    """Iterate over type_infos such that each type_info precedes its sub types."""
    handled_global_ids = set()
    while type_infos:
        infos_to_handle = {}
        progress_made = False
        for global_id, info in type_infos.items():
            if info.global_super_type_ids.issubset(handled_global_ids):
                progress_made = True
                handled_global_ids.add(global_id)
                yield info
            else:
                infos_to_handle[global_id] = info
        assert progress_made, 'no progress made (missing ids? cycle?)'
        type_infos = infos_to_handle

class XmlImporter:
    """
    loader: callable with one arg (uri) that returns the XML root element corresponding to
        the model with the given uri
    """

    def __init__(self, loader, uris, metamodel=None):
        if not metamodel:
            from pnb.mcl.metamodel import standard as metamodel

        self.metamodel = metamodel
        self.model_info_by_uri = {}
        self.model_by_uri = {}
        self.type_info_by_global_id = {}

        self.built_type_by_global_id = {}
        self.built_object_infos = []
        self.built_object_by_global_id = {}
        
        if isinstance(uris, str):
            self.uris_to_handle = {uris}
        else:
            self.uris_to_handle = set(uris)
            

        # import all xml models
        while self.uris_to_handle:
            uri = self.uris_to_handle.pop()
            assert uri not in self.model_info_by_uri
            self.model_info_by_uri[uri] = self._get_model_infos(loader, uri)

        # generate type infos from all xml models
        for model_info in self.model_info_by_uri.values():
            self._generate_type_infos(model_info.xml_element)

        # build mcl types from sorted type infos
        sorted_type_infos = list(sorted_by_dependency(self.type_info_by_global_id))
        for info in sorted_type_infos:
            self._build_type(info)

        # build mcl properties for types
        for global_id, type_info in self.type_info_by_global_id.items():
            self._build_type_properties(global_id, type_info.xml_element)

        # for each xml model: build mcl model with all sub packages and packaged elements
        for model_info in self.model_info_by_uri.values():
            model = self._build_package(model_info.xml_element)
            self.model_by_uri[model.uri] = model

        # build mcl properties for objects
        for object_info in self.built_object_infos:
            self._build_object_properties(
                object_info.built_object, object_info.xml_element)
            
        if isinstance(uris, str):

            self.model = self.model_by_uri[uris]

    def _get_model_infos(self, loader, uri):
        xml_element = loader(uri)
        import_uris_by_prefix = {}

        for xml_import in xml_element.iterchildren('Import'):
            import_uri = xml_import.attrib.get('source')
            assert import_uri != uri
            if import_uri not in self.model_info_by_uri:
                self.uris_to_handle.add(import_uri)
            import_prefix = xml_import.attrib.get('prefix')
            assert import_prefix is not None, xml_import
            assert import_prefix not in import_uris_by_prefix
            import_uris_by_prefix[import_prefix] = import_uri

        return ModelInfo(uri, xml_element, import_uris_by_prefix)

    def _get_global_item_ids(self, xml_element, xml_attribute):
        global_item_ids = []
        item_ids_str = xml_element.attrib.get(xml_attribute)
        if item_ids_str is not None:
            item_ids = item_ids_str.split()
        else:
            return set()

        xml_parent = xml_element.getparent()
        while xml_parent.tag != 'Model':
            xml_parent = xml_parent.getparent()
        xml_model_uri = xml_parent.attrib.get('uri')

        for item_id in item_ids:
            if item_id.startswith('builtin'):
                global_item_ids.append(item_id)
                continue
            if item_id.startswith(URI_SEP_SYMBOL):
                global_item_id = xml_model_uri + item_id
            elif item_id.startswith(ID_SEP_SYMBOL):
                global_item_id = xml_model_uri + URI_SEP_SYMBOL + item_id[1:]
            else:
                prefix = item_id.split(URI_SEP_SYMBOL)[0]
                xml_model_info = self.model_info_by_uri[xml_model_uri]
                import_uri = xml_model_info.import_uris_by_prefix[prefix]
                suffix = item_id[len(prefix):]
                global_item_id = import_uri + suffix

            global_item_ids.append(global_item_id)

        return global_item_ids

    def _get_xml_id_and_sep_symbol(self, xml_element, parent_id):
        xml_id = xml_element.attrib.get('name')
        sep_symbol = SUB_ELEMENT_SEP_SYMBOL
        if xml_id is None:
            xml_id = xml_element.attrib.get('id')
        if parent_id is not None and parent_id[-1] == URI_SEP_SYMBOL:
            sep_symbol = ''
        return xml_id, sep_symbol

    def _get_converted_data_values(self, xml_data_element):
        converted_values = []
        for child_element in xml_data_element:
            value_str = child_element.text
            match child_element.tag:
                case 'String':
                    converted_values.append(value_str)
                case 'Boolean':
                    converted_values.append(bool(value_str))
                case 'Double':
                    converted_values.append(float(value_str))
                case 'Int':
                    converted_values.append(int(value_str))
                case 'DateTime':
                    # fromisoformat should work adequately for python >= 3.11
                    converted_values.append(datetime.fromisoformat(value_str))

        return converted_values

    def _generate_type_infos(self, xml_element, parent_id=None):
        xml_id, sep_symbol = self._get_xml_id_and_sep_symbol(xml_element, parent_id)

        match xml_element.tag:
            case 'Model':
                uri = xml_element.attrib.get('uri')
                assert uri is not None
                assert xml_id is not None
                for child_element in xml_element.iterchildren():
                    self._generate_type_infos(child_element, uri + URI_SEP_SYMBOL)
            case 'Package':
                assert xml_id is not None
                for child_element in xml_element.iterchildren():
                    self._generate_type_infos(
                        child_element,
                        parent_id + sep_symbol + xml_id)
            case 'Import':
                return
            case 'Object':
                return
            case _:
                assert xml_element.tag in ALL_TYPES, xml_element.tag
                global_id = parent_id + sep_symbol + xml_id
                assert global_id not in self.type_info_by_global_id
                assert xml_id is not None
                super_type_xml_ids = self._get_global_item_ids(xml_element, 'superTypes')
                self.type_info_by_global_id[global_id] = TypeInfo(
                    global_id, set(super_type_xml_ids), xml_element)

    def _build_type(self, type_info):
        xml_element = type_info.xml_element
        xml_tag = xml_element.tag
        name = xml_element.attrib['name']
        super_types = [self.built_type_by_global_id[id] for id in type_info.global_super_type_ids]

        assert xml_tag in ALL_TYPES
        meta_type = getattr(self.metamodel, xml_tag)

        assert type_info.global_id not in self.built_type_by_global_id
        self.built_type_by_global_id[type_info.global_id] = meta_type(name, super_types)

        return

    def _build_type_properties(self, global_id, xml_element):
        for xml_prop in xml_element:
            
            if xml_prop.tag not in ('CompositionProperty', 'DataProperty', 'ReferenceProperty'):
                continue
            
            name = xml_prop.attrib.get('name')
            is_ordered_str = xml_prop.attrib.get('isOrdered')
            if is_ordered_str == 'false':
                is_ordered = False
            else:
                is_ordered = True
            print(xml_prop)
            lower = int(xml_prop.attrib.get('lower'))
            global_type_id = self._get_global_item_ids(xml_prop, 'type').pop()
            if global_type_id.startswith('builtin'):
                type_ = getattr(self.metamodel, global_type_id[len('builtin')+1:])
            else:
                type_ = self.built_type_by_global_id[global_type_id]
            upper_str = xml_prop.attrib.get('upper')
            if upper_str == '*':
                upper = None
            else:
                upper = int(upper_str)

            match xml_prop.tag:
                case 'CompositionProperty':
                    self.built_type_by_global_id[global_id].ownedAttributes.add(
                        self.metamodel.CompositionProperty(
                            name, type_, lower, upper, is_ordered))
                case 'DataProperty':
                    is_unique_str = xml_prop.attrib.get('isUnique')
                    if is_unique_str == 'false':
                        is_unique = False
                    else:
                        is_unique = True
                    self.built_type_by_global_id[global_id].ownedAttributes.add(
                        self.metamodel.DataProperty(
                            name, type_, lower, upper, is_ordered, is_unique))
                case 'ReferenceProperty':
                    is_unique_str = xml_prop.attrib.get('isUnique')
                    if is_unique_str == 'false':
                        is_unique = False
                    else:
                        is_unique = True
                    opp_lower = int(xml_prop.attrib.get('oppositeLower'))
                    opp_upper_str = xml_prop.attrib.get('oppositeUpper')
                    if opp_upper_str == '*':
                        opp_upper = None
                    else:
                        opp_upper = int(opp_upper_str)
                    self.built_type_by_global_id[global_id].ownedAttributes.add(
                        self.metamodel.ReferenceProperty(
                            name, type_, lower, upper, is_ordered, is_unique,
                            opp_lower, opp_upper))
        return

    def _build_object(self, xml_element, parent_id):
        # build mcl object with all sub objects
        global_type_id = self._get_global_item_ids(xml_element, 'type').pop()
        type_ = self.built_type_by_global_id[global_type_id]
        name = xml_element.attrib.get('name')

        object_ = self.metamodel.Object(type_, name)
        object_info = ObjectInfo(object_, xml_element)
        self.built_object_infos.append(object_info)

        # check whether object has id for referencing
        xml_id, sep_symbol = self._get_xml_id_and_sep_symbol(xml_element, parent_id)
        if xml_id is not None:
            global_object_id = parent_id + sep_symbol + xml_id
            self.built_object_by_global_id[global_object_id] = object_
            child_parent_id = global_object_id
        else:
            child_parent_id = parent_id

        # check for sub objects
        for child_element in xml_element:
            if child_element.tag == 'Components':
                # composition property
                prop_name = child_element.attrib.get('property')
                prop = type_.attributes.get(prop_name)
                assert isinstance(prop, self.metamodel.CompositionProperty), prop
                values = []
                for sub_element in child_element:
                    assert sub_element.tag == 'Object'
                    values.append(self._build_object(sub_element, child_parent_id))
                prop._set_values_(object_, values)
            else:
                # data and reference properties
                pass

        return object_

    def _build_object_properties(self, built_object, xml_element):
        global_type_id = self._get_global_item_ids(xml_element, 'type').pop()
        type_ = self.built_type_by_global_id[global_type_id]
        for child_element in xml_element:
            match child_element.tag:
                case 'Components':
                    # sub objects already built
                    pass
                case 'Data':
                    prop_name = child_element.attrib.get('property')
                    prop = type_.attributes.get(prop_name)
                    assert isinstance(prop, self.metamodel.DataProperty), prop
                    values = self._get_converted_data_values(child_element)
                    prop._set_values_(built_object, values)
                case 'References':
                    prop_name = child_element.attrib.get('property')
                    prop = type_.attributes.get(prop_name)
                    assert isinstance(prop, self.metamodel.ReferenceProperty), prop
                    values = []
                    global_object_ids = self._get_global_item_ids(child_element, 'refs')
                    for global_object_id in global_object_ids:
                        values.append(self.built_object_by_global_id[global_object_id])
                    prop._set_values_(built_object, values)
                case _:
                    raise TypeError(child_element)

        return

    def _build_package(self, xml_package_element, parent_id=None):
        # build mcl package with all sub packages and all packaged elements
        # use already built mcl types
        name = xml_package_element.attrib.get('name')
        xml_tag = xml_package_element.tag

        if xml_tag == 'Model':
            uri = xml_package_element.attrib.get('uri')
            package = self.metamodel.Model(name, uri)
            assert parent_id is None
            parent_id = uri + URI_SEP_SYMBOL
        else:
            assert xml_tag == 'Package'
            package = self.metamodel.Package(name)

        for child_element in xml_package_element:
            match child_element.tag:
                case 'Package':
                    xml_id, sep_symbol = self._get_xml_id_and_sep_symbol(child_element, parent_id)
                    sub_package = self._build_package(
                        child_element,
                        parent_id + sep_symbol + xml_id)
                    package.packagedElements.add(sub_package)
                case type_ if type_ in ALL_TYPES:
                    xml_id, sep_symbol = self._get_xml_id_and_sep_symbol(child_element, parent_id)
                    global_id = parent_id + sep_symbol + xml_id
                    built_type = self.built_type_by_global_id[global_id]
                    package.packagedElements.add(built_type)
                case 'Object':
                    object_ = self._build_object(child_element, parent_id)
                    package.add(object_)
                case _:
                    assert child_element.tag == 'Import'
                    continue

        return package

def read_xml(path, uri):
    reader = XmlImporter(path, uri)
    return reader.model


def read_xmls(xml_paths):
    
    xml_root_by_uri = {}
    for path in xml_paths:
        xml_root = etree.parse(path).getroot()
        uris = xml_root.xpath('/Model/@uri')
        assert len(uris) == 1, uris
        uri = uris[0]
        assert uri not in xml_root_by_uri
        xml_root_by_uri[uri] = xml_root
        
    importer = XmlImporter(xml_root_by_uri.get, xml_root_by_uri)

    model_by_name = {}
    for model in importer.model_by_uri.values():
        assert model.name not in model_by_name
        model_by_name[model.name] = model

    return model_by_name




from unittest import TestCase
from lxml import etree
from pnb.mcl.metamodel import standard
from pnb.mcl.io.xml import XmlExporter
from pnb.mcl.test.examples import ex_1

def normalize_xml_code(code):
    return ''.join(row.strip() for row in code.splitlines())

class Test_XmlExporter_ex_1(TestCase):

    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        cls.model_by_uri = ex_1.make_model_by_uri(standard)

    def test_ModelCore(self):
        exporter = XmlExporter(self.model_by_uri['http://ModelCore'])
        self.assertEqual(
            etree.tostring(exporter.xml, encoding='unicode'),
            normalize_xml_code('''
                <Model name="ModelCore" uri="http://ModelCore">
                  <AbstractClass name="ConceptualModel"/>
                </Model>'''))

    def test_ModelProcess(self):
        exporter = XmlExporter(self.model_by_uri['http://ModelProcess'])
        self.assertEqual(
            etree.tostring(exporter.xml, encoding='unicode'),
            normalize_xml_code('''
                <Model name="ModelProcess" uri="http://ModelProcess">
                  <Import source="http://ModelCore" prefix="ModelCore"/>
                  <Package name="Base">
                    <ConcreteClass name="ProcessStep" superTypes="ModelCore/ConceptualModel">
                      <ReferenceProperty name="RelatedSteps" isOrdered="false" isUnique="true" lower="0" oppositeLower="0" oppositeUpper="*" type="/Base.ProcessStep" upper="*"/>
                      <DataProperty name="StepNumber" isOrdered="false" isUnique="false" lower="0" type="builtin/String" upper="1"/>
                      <CompositionProperty name="SubSteps" isOrdered="false" lower="0" type="/Base.ProcessStep" upper="*"/>
                    </ConcreteClass>
                  </Package>
                  <Package name="Steps">
                    <ConcreteClass name="Pumping" superTypes="/Base.ProcessStep"/>
                  </Package>
                </Model>'''))

    def test_ModelInstance1(self):
        exporter = XmlExporter(self.model_by_uri['http://ModelInstance1'])
        self.assertEqual(
            etree.tostring(exporter.xml, encoding='unicode'),
            normalize_xml_code('''
                <Model name="ModelInstance1" uri="http://ModelInstance1">
                  <Import source="http://ModelInstance2b" prefix="ModelInstance20"/>
                  <Import source="http://ModelInstance2a" prefix="ModelInstance2"/>
                  <Import source="http://ModelProcess" prefix="ModelProcess"/>
                  <Object type="ModelProcess/Steps.Pumping">
                    <References refs="ModelInstance2/Step1 ModelInstance20/Step" property="RelatedSteps"/>
                    <Data property="StepNumber">
                      <String>step1</String>
                    </Data>
                  </Object>
                </Model>'''))

    def test_ModelInstance2a(self):
        exporter = XmlExporter(self.model_by_uri['http://ModelInstance2a'])
        self.assertEqual(
            etree.tostring(exporter.xml, encoding='unicode'),
            normalize_xml_code('''
                <Model name="ModelInstance2" uri="http://ModelInstance2a">
                  <Import source="http://ModelInstance3" prefix="ModelInstance3"/>
                  <Import source="http://ModelProcess" prefix="ModelProcess"/>
                  <Object name="Step1" type="ModelProcess/Steps.Pumping">
                    <References refs="#Pumping1 ModelInstance3/Step" property="RelatedSteps"/>
                    <Data property="StepNumber">
                      <String>step2a1</String>
                    </Data>
                  </Object>
                  <Object id="Pumping1" type="ModelProcess/Steps.Pumping">
                    <Data property="StepNumber">
                      <String>step2a2</String>
                    </Data>
                  </Object>
                </Model>'''))

    def test_ModelInstance2b(self):
        exporter = XmlExporter(self.model_by_uri['http://ModelInstance2b'])
        self.assertEqual(
            etree.tostring(exporter.xml, encoding='unicode'),
            normalize_xml_code('''
                <Model name="ModelInstance2" uri="http://ModelInstance2b">
                  <Import source="http://ModelInstance2a" prefix="ModelInstance2"/>
                  <Import source="http://ModelProcess" prefix="ModelProcess"/>
                  <Object name="Step" type="ModelProcess/Steps.Pumping">
                    <References refs="ModelInstance2/Step1" property="RelatedSteps"/>
                    <Data property="StepNumber">
                      <String>step2b</String>
                    </Data>
                  </Object>
                </Model>'''))

    def test_ModelInstance3(self):
        exporter = XmlExporter(self.model_by_uri['http://ModelInstance3'])
        self.assertEqual(
            etree.tostring(exporter.xml, encoding='unicode'),
            normalize_xml_code('''
                <Model name="ModelInstance3" uri="http://ModelInstance3">
                  <Import source="http://ModelProcess" prefix="ModelProcess"/>
                  <Object type="ModelProcess/Steps.Pumping">
                    <Data property="StepNumber">
                      <String>step31</String>
                    </Data>
                    <Components property="SubSteps">
                      <Object name="Step" type="ModelProcess/Steps.Pumping">
                        <Data property="StepNumber">
                          <String>step32</String>
                        </Data>
                      </Object>
                    </Components>
                  </Object>
                </Model>'''))



import itertools
import sys

import rdflib
from rdflib import Literal, RDF, BNode, OWL, RDFS, XSD


class OwlExporter:

    def __init__(self, model, prefix=None, uri=None):
        
        self.redundantSubClassAxioms = True
        self.redundantTypeAxioms = True
        self.abstractSuperProperties = True
        self.topLevelTypes = True
        
        self.model = model
        self.namespace = rdflib.Namespace(uri or (model.uri + '#'))
        self.metamodel = sys.modules[model.__module__]
        self.graph = rdflib.Graph()
        self.graph.bind(prefix or self.model.name, self.namespace)
        self.node_by_element = {}
        self.element_by_node = {}
        self.defined_nodes = set()
        
        self.graph.add((rdflib.URIRef(str(self.namespace)[:-1]), RDF.type, OWL.Ontology))
        
        self.value_property = self.namespace.value
        self.graph += [
            (self.value_property, RDF.type, OWL.DatatypeProperty),
            (self.value_property, RDF.type, OWL.FunctionalProperty)]

        if self.abstractSuperProperties:
            
            self.composition_property = self.namespace.CompositionProperty
            self.graph += [
                (self.composition_property, RDF.type, OWL.ObjectProperty)]
            
            self.reference_property = self.namespace.ReferenceProperty
            self.graph += [
                (self.reference_property, RDF.type, OWL.ObjectProperty)]
            
            self.data_property = self.namespace.DataProperty
            self.graph += [
                (self.data_property, RDF.type, OWL.ObjectProperty)]

        self.add_element(model)

        undefined_nodes = set(self.element_by_node).difference(self.defined_nodes)
        assert not undefined_nodes, undefined_nodes
        
        top_level_classes = [
            element for element in self.node_by_element
            if isinstance(element, self.metamodel.Class) and not element.superTypes]
        
        self._add_disjointness_if_applicable(top_level_classes, 'top_level_classes')

        top_level_datatypes = [
            element for element in self.node_by_element
            if isinstance(element, self.metamodel.DataType) and not element.superTypes]
        
        self._add_disjointness_if_applicable(top_level_datatypes, 'top_level_datatypes')
        
        if self.topLevelTypes:
            
            self.graph += [
                (self.namespace.Object, RDF.type, OWL.Class)]
            self.graph += equivalent_union(self.namespace.Object, [self.get_node(c) for c in top_level_classes])
            
            if self.redundantSubClassAxioms:
                for c in top_level_classes:
                    self.graph.add((self.get_node(c), RDFS.subClassOf, self.namespace.Object))  

            self.graph += [
                (self.namespace.DataValue, RDF.type, OWL.Class)]
            self.graph += equivalent_union(self.namespace.DataValue, [self.get_node(c) for c in top_level_datatypes])
            
            if self.redundantSubClassAxioms:
                for c in top_level_datatypes:
                    self.graph.add((self.get_node(c), RDFS.subClassOf, self.namespace.DataValue))
    
            self.graph += all_disjoint_classes((self.namespace.Object,  self.namespace.DataValue))
            
            
            if self.abstractSuperProperties:
                self.graph += [
                    (self.composition_property, RDFS.domain, self.namespace.Object),
                    (self.composition_property, RDFS.range, self.namespace.Object)]
                
                self.graph += [
                    (self.reference_property, RDFS.domain, self.namespace.Object),
                    (self.reference_property, RDFS.range, self.namespace.Object)]
                
                
                object_or_data_value = union_of((self.namespace.Object, self.namespace.DataValue))
                self.graph += object_or_data_value
                
                self.graph += [
                    (self.data_property, RDFS.domain, object_or_data_value.class_),
                    (self.data_property, RDFS.range, self.namespace.DataValue)]
                    

                
                
                
                


    def get_node(self, element):
        node = self.node_by_element.get(element)
        if node is None:
            if isinstance(element, (self.metamodel.Type, self.metamodel.SingletonValue)):
                name = element.name
            elif isinstance(element, (self.metamodel.EnumerationLiteral, self.metamodel.Property)):
                name = element.owner.name + '.' + element.name
            else:
                raise Exception(element)
            node = self.namespace.term(name)
            self.node_by_element[element] = node
            assert node not in self.element_by_node
            self.element_by_node[node] = element
        return node

    def def_node(self, element, owl_type):
        node = self.get_node(element)
        assert node not in self.defined_nodes
        self.graph.add((node, RDF.type, owl_type))
        self.defined_nodes.add(node)
        return node

    def add_element(self, element):
        if isinstance(element, self.metamodel.Model):
            for pe in element.packagedElements:
                self.add_element(pe)
                # TODO: unnamed objects
        elif isinstance(element, self.metamodel.Package):
            for pe in element.packagedElements:
                self.add_element(pe)
        elif isinstance(element, self.metamodel.Type):
            self.add_type(element)
        else:
            pass
        
        
    def _add_disjointness_if_applicable(self, types, info):
        if len(types) < 2:
            return
        
        types = [type_ for type_ in types if type_.name not in ['ColumnSection', 
                                                                'Vent']]
                                                                
                                  
        if info == 'top_level_classes':      
                                                                
            types = [type_ for type_ in types if type_.name not in [
                                                                
                                                                'CustomAttributeOwner',
                                                                'CustomObject',
                                                                'ChamberOwner',
                                                                'NozzleOwner',
                                                                'PipingNodeOwner',
                                                                'TransmissionDriver',
                                                                'ActuatingElectricalLocation',
                                                                'SensingLocation',
                                                                'SignalConveyingFunctionSource',
                                                                'SignalConveyingFunctionTarget',
                                                                'PipingConnection',
                                                                'PipingSourceItem',
                                                                'PipingTargetItem',
                                                                'PipingNetworkSegmentItem',
                                                                'IndustrialComplexParentStructure',
                                                                'PlantAreaLocatedStructure',
                                                                'PlantSectionParentStructure',
                                                                'PlantSystemLocatedStructure',
                                                                'PlantTrainLocatedStructure',
                                                                'ProcessPlantParentStructure',
                                                                'TechnicalItemParentStructure',
                                                                ]]
        
        
        
        applicable = True
        common_subtypes = set()
        print('########################')
        for type_ in types:
            print('+++', type_)
            if type_ in common_subtypes:
                applicable = False
                break
            common_subtypes.add(type_)
            new_subtypes = type_.allSubTypes
            if not common_subtypes.isdisjoint(new_subtypes):
                applicable = False
                break
            common_subtypes.update(new_subtypes)
        if not applicable:
            print(common_subtypes.intersection(new_subtypes))
            print(f'### {info} {", ".join(sorted(t.name for t in types))} not disjoint')
            assert False
            return
        
        self.graph += all_disjoint_classes(self.get_node(type_) for type_ in types)


        
    def add_type(self, type_):

        node = self.def_node(type_, OWL.Class)
        subTypes = list(type_.subTypes)
        
        self._add_disjointness_if_applicable(subTypes, type_)
        
        subTypeNodes = [self.get_node(subType) for subType in subTypes]
        
        if self.redundantSubClassAxioms or not type_.isAbstract:
            for subTypeNode in subTypeNodes:
                self.graph.add((subTypeNode, RDFS.subClassOf, node))
        
        if type_.isAbstract:
            self.graph += equivalent_union(node, subTypeNodes)

        if isinstance(type_, self.metamodel.PrimitiveType):
            
            name = type_.name
            xsd_name = name[0].lower() + name[1:] 
            xsd_node = getattr(XSD, xsd_name)
            self.graph += restrict_all_values_from(node, self.value_property, OWL.DatatypeProperty, xsd_node)
            self.graph += restrict_min_cardinality(node, self.value_property, OWL.DatatypeProperty, 1)

            
        if isinstance(type_, self.metamodel.Enumeration):
            literal_nodes = [self.def_node(literal, OWL.NamedIndividual) for literal in type_.orderedOwnedLiterals]
            assert len(literal_nodes), type_
            if not literal_nodes:
                ERROR
            if self.redundantTypeAxioms:
                for literal_node in literal_nodes:
                    self.graph.add((literal_node, RDF.type, node))
            self.graph += equivalent_one_of(node, literal_nodes)
            if len(literal_nodes) > 1:
                self.graph += all_different(literal_nodes)
                
        if isinstance(type_, self.metamodel.SingletonType):
            value = type_.value
            assert value
            value_node = self.def_node(value, OWL.NamedIndividual)
            if self.redundantTypeAxioms:
                self.graph.add((value_node, RDF.type, node))
            self.graph += equivalent_one_of(node, [value_node])
            
            
            
            
            
            
            
        if type_.name == 'CustomAttribute':
            value_prop_node = self.namespace.term('CustomAttribute.Value')
            self.graph += [
                (value_prop_node, RDF.type, OWL.ObjectProperty),
                (value_prop_node, RDFS.domain, node)]

            self.graph += restrict_cardinality(node, value_prop_node, OWL.ObjectProperty, 1)

            if self.abstractSuperProperties:
                self.graph.add((value_prop_node, RDFS.subPropertyOf, self.data_property))
            
            
            
            
        for prop in type_.ownedAttributes:
            
            is_ca_value = False
            if prop.name == 'Value' and type_.name.startswith('Custom'):
                assert len(list(type_.superTypes)) == 1, prop
                if list(type_.superTypes)[0].name == 'CustomAttribute':
                    is_ca_value = True
                
                
                
            if not is_ca_value:
                
                           
            
            
                prop_node = self.def_node(prop, OWL.ObjectProperty)
                
                self.graph += [
                    (prop_node, RDFS.domain, node),
                    (prop_node, RDFS.range, self.get_node(prop.type))]
                
                if self.abstractSuperProperties:
                    super_prop_node = {
                        self.metamodel.CompositionProperty: self.composition_property,
                        self.metamodel.ReferenceProperty: self.reference_property,
                        self.metamodel.DataProperty: self.data_property}[type(prop)]
                    self.graph.add((prop_node, RDFS.subPropertyOf, super_prop_node))
                    
                if prop.lower == prop.upper:
                    self.graph += restrict_cardinality(node, prop_node, OWL.ObjectProperty, prop.lower, False)
                    
                else:
                    if prop.lower > 0:
                        self.graph += restrict_min_cardinality(node, prop_node, OWL.ObjectProperty, prop.lower, False)
                    if prop.upper is not None:
                        self.graph += restrict_max_cardinality(node, prop_node, OWL.ObjectProperty, prop.upper, False)
                    
            else:
                prop_node = self.namespace.term('CustomAttribute.Value')
                self.graph += restrict_all_values_from(node, prop_node, OWL.ObjectProperty, self.get_node(prop.type))
                

                
                
                

                    
      

            
            

          #  ji
            
            
                
                
                
                


    


                





"""
:todo Doc:
"""


__docformat__ = 'restructuredtext en'



import itertools
import operator

import six



import rdflib
from rdflib import OWL, RDF, RDFS, XSD

#import rdfutils
#from rdfutils import owlutils
#from attrs_utils import define



BASE = rdflib.Namespace('http://www.aixcape.org/oimel/base.owl#')



def paragraphs(text):
    
    
    pars = []
    par_lines = []
    for line in text.split('\n'):
        line = line.strip()
        if line:
            par_lines.append(line)
        elif par_lines:
            pars.append(' '.join(par_lines))
            par_lines = []
    if par_lines:
        pars.append(' '.join(par_lines))
    return '\n'.join(pars)


class Uml2OwlConversion(object):
    """:todo Doc:"""

    def __init__(self):
        self.owl_graph = rdflib.ConjunctiveGraph()
        self.uri_by_element = {}
        self._element_by_uri = {}
        self.graph_by_package = {}
        self.infos_by_element = {}
        self.warnings_by_element = {}
        
        self.uml_subtypes_by_uml_type = {}
        
        self._all_uml_subtypes_by_uml_type = {}
        
        
    def arrange_types(self):
        
        all_subtypes = set()
        
        for element in self._element_by_uri.values():
            if isinstance(element, uml.Type):
                all_subtypes.update(self.get_all_uml_subtypes(element).difference((element, )))
                
        self.top_level_types = set(
            element for element in self._element_by_uri.values()
            if isinstance(element, uml.Type))
        self.top_level_types.difference_update(all_subtypes)

        
    def get_all_uml_subtypes(self, type_):
        
        result = self._all_uml_subtypes_by_uml_type.get(type_)
        if result is None:
            result = set((type_, ))
            for subtype in self.uml_subtypes_by_uml_type.get(type_, []):
                result.update(self.get_all_uml_subtypes(subtype))
            result = frozenset(result)
            self._all_uml_subtypes_by_uml_type[type_] = result
            
        return result
        
        


    def add_warning(self, element, message, formatter=paragraphs):
        """:todo Doc:"""
        self.warnings_by_element.setdefault(element, []).append(
            formatter(message))
  
    def add_info(self, element, message, formatter=paragraphs):
        assert isinstance(element, uml.Element)
        assert isinstance(message, six.text_type)
        self.infos_by_element.setdefault(element, []).append(
            formatter(message))

        
        
    def register_element(self, element, uri):
        """:todo Doc:"""
        
        assert isinstance(element, uml.Element)
        if element in self.uri_by_element:
            ERROR        
        if uri is not None:
            assert isinstance(uri, rdflib.URIRef)
            if not rdfutils.is_uriref(uri):
                ERROR
            if uri in self._element_by_uri:
                assert uri not in self._element_by_uri, uri
                ERROR

        self.uri_by_element[element] = uri
        if uri is not None:
            self._element_by_uri[uri] = element


    def get_element_uri(self, element, required=True):
        try:
            result = self.uri_by_element[element]
            if required and result is None:
                ERRIR
            return result
        except KeyError:
            if required:
                raise
            else:
                return None      
            

    def get_element_uris(self, elements):
        return [self.get_element_uri(element) for element in elements]
                  

    
# TODO begin final


class Uml2OwlConverterConfig:
    local_enumeration_literal_uris: bool=False
    merge_packages: bool=False
    
    

class Uml2OwlConverter(object):
    """:todo Doc:"""

    def __init__(self, config=Uml2OwlConverterConfig()):
        """:todo Doc:"""
        
        self.config = config

        self._owl_ext = '.owl'
        self._owl_sep = '#'
        self._owl_type_by_uml_primitive_type = {
            uml.Boolean: XSD.boolean,
            uml.String: XSD.string,
            uml.Integer: XSD.integer,
            uml.Real: XSD.decimal}


    def run(self, package, conversion=None):
        """:todo Doc:"""
        
        

        if not conversion:
            conversion = Uml2OwlConversion()


        
        self._convert_types(package, conversion)
        self._convert_generalizations(package, conversion)
        self._add_generalization_axioms(package, conversion)
        
        self._convert_enumeration_literals(package, conversion)
        self._convert_instances(package, conversion)
        
        self._complete_primitives(package, conversion)
        if 0:
        
            self._convert_enumerations(package, conversion)
            
            self._convert_primitive_types(package, conversion)
            self._convert_generalizations(package, conversion)
            self._convert_generalization_sets(package, conversion)
            
            
            
            
        self._convert_properties(package, conversion)
            
        
        

        return conversion
    
# TODO end final




    def _add_generalization_axioms(self, root_package, conversion):
        conversion.arrange_types()
        for supertype, subtypes in itertools.chain(
                conversion.uml_subtypes_by_uml_type.items(),
                [(None, conversion.top_level_types)]):

            subtype_uris = {
                conversion.get_element_uri(subtype)
                for subtype in subtypes}
            graph = self._get_graph_for_type(
                root_package, conversion, supertype)

            if supertype and supertype.uml_isAbstract:
                supertype_uri = conversion.get_element_uri(supertype)
                graph += owlutils.equivalent_union(supertype_uri, subtype_uris) 
            if len(subtypes) > 1:
                indirect_subtypes = set()
                for subtype in subtypes:
                    new_subtypes = conversion.get_all_uml_subtypes(subtype)
                    if not indirect_subtypes.isdisjoint(new_subtypes):
                        info = indirect_subtypes.intersection(new_subtypes)
                        info = ' from '+ subtype.uml_name + ': ' + ', '.join(cls.uml_name for cls in info) + ' from '+ subtype.uml_name 
                        print(f'subtypes of {supertype.uml_name if supertype else "TOP"} not disjoint: {info}')
                        break
                    indirect_subtypes.update(new_subtypes)
                else:
                    graph += owlutils.all_disjoint_classes(subtype_uris)



    
    


    
# TODO begin final




    def _convert_enumeration_literals_of_enumeration(
            self, root_package, conversion, enumeration):
        # pylint: disable=invalid-name
        """:todo Doc:"""

        if enumeration.uml_isAbstract:
            if enumeration.uml_ownedLiteral:
                conversion.add_warning(
                    enumeration,
                    '''The EnumerationLiterals of this Enumeration have been
                        skipped because the Enumeration is abstract.''')
            return

        if not enumeration.uml_ownedLiteral:
            conversion.add_warning(
                enumeration,
                '''For this non-abstract Enumeration, no EnumerationLiterals are
                    given.''')
            return

        enumeration_uri = conversion.get_element_uri(enumeration)
        literal_uris = []

        for literal in enumeration.uml_ownedLiteral:
            literal_uri = self._create_uri_for_enumeration_literal(
                root_package, conversion, literal)
            literal_uris.append(literal_uri)
            conversion.register_element(literal, literal_uri)
            literal_graph = self._get_graph_for_enumeration_literal(
                root_package, conversion, literal)
            literal_graph += owlutils.a_named_individual(
                literal_uri, enumeration_uri)

        enumeration_graph = self._get_graph_for_type(
            root_package, conversion, enumeration)
        enumeration_graph += owlutils.equivalent_one_of(
            enumeration_uri, literal_uris)
        if len(literal_uris) > 1:
            enumeration_graph += owlutils.all_different(literal_uris)


    def _create_uri_for_enumeration_literal(
            self, root_package, conversion, enumeration_literal):
        """:todo Doc:"""
        
        namespace = self._get_graph_for_enumeration_literal(
            root_package, conversion, enumeration_literal).identifier
            
        if self.config.local_enumeration_literal_uris:
            enumeration = enumeration_literal.uml_owner
            return self._create_uri(
                f'{enumeration.uml_name}.{enumeration_literal.uml_name}',
                namespace)
        else:
            return self._create_uri_from_name(
                root_package, conversion, enumeration_literal, namespace)

    def _get_graph_for_enumeration_literal(
            self, root_package, conversion, enumeration_literal):
        """:todo Doc:"""

        package = enumeration_literal.uml_enumeration.uml_package
        return self._get_graph_for_package(root_package, conversion, package)
    
    
    def _convert_instances(self, root_package, conversion):
        """:todo Doc:"""

        for nested_package in self._iter_nested_packages(root_package):
            for packaged_element in nested_package.uml_packagedElement:
                if type_is(packaged_element, uml.InstanceSpecification):
                    self._convert_instance(
                        root_package, conversion, packaged_element)
                    
                    
    def _convert_instance(self, root_package, conversion, instance):
        # pylint: disable=invalid-name
        """:todo Doc:"""

        package = instance.uml_owner
        namespace = self._get_graph_for_package(root_package, conversion, package).identifier
        uri = self._create_uri(instance.uml_name, namespace)
        conversion.register_element(instance, uri)
        graph = self._get_graph_for_package(root_package, conversion, package)
        graph += owlutils.a_named_individual(uri)

        classifiers = instance.uml_classifier
        assert len(classifiers) == 1

        classifier = set(classifiers).pop()
        classifier_uri = conversion.get_element_uri(classifier)

        graph += [(uri, RDF.type, classifier_uri)]
        graph += owlutils.equivalent_one_of(classifier_uri, [uri])

    def _convert_generalizations(self, root_package, conversion):
        """:todo Doc:"""

        for nested_package in self._iter_nested_packages(root_package):
            for packaged_element in nested_package.uml_packagedElement:
                if isinstance(packaged_element, uml.Type):
                    if not isinstance(packaged_element, uml.Association):
                        self._convert_generalizations_of_type(
                            root_package, conversion, packaged_element)
                        
                        
                        
                        
    def _complete_primitives(self, root_package, conversion):
        """:todo Doc:"""

        for nested_package in self._iter_nested_packages(root_package):
            for packaged_element in nested_package.uml_packagedElement:
                if type_is(packaged_element, uml.PrimitiveType):
                    primitive_type = packaged_element
                    
                    graph = self._get_graph_for_type(
                        root_package, conversion, primitive_type)
                    
                    value_uri = self._create_uri('value', graph.identifier)
                    
                    primitive_type_uri = conversion.get_element_uri(primitive_type)
                    
                    graph += [
                        (value_uri, RDF.type, OWL.DatatypeProperty),
                        (value_uri, RDF.type, OWL.FunctionalProperty)]
                    
                    
                    type_uri = dict(
                        AnyURI=XSD.anyURI,
                        Boolean=XSD.boolean,
                        DateTime=XSD.dateTime,
                        Double=XSD.double,
                        Integer=XSD.integer,
                        String=XSD.string,
                        UnsignedByte=XSD.unsignedByte)[primitive_type.uml_name]
                    
                    
                    graph += owlutils.restrict_all_values_from(
                        primitive_type_uri, value_uri, OWL.DatatypeProperty, type_uri)
                    graph += owlutils.restrict_min_cardinality(
                        primitive_type_uri, value_uri, OWL.DatatypeProperty, 1)
                        
                        
                        
                  
                    
                    
                    
                    
                    
                    
                    

                    
 
                        
                        
                        
                        
                        
                        
                        

    def _convert_generalizations_of_type(
            self, root_package, conversion, class_):
        """:todo Doc:"""

        class_uri = conversion.get_element_uri(class_)

        for generalization in class_.uml_generalization:
            general = generalization.uml_general
            general_uri = conversion.get_element_uri(
                general, required=False)
            if general_uri:
                graph = self._get_graph_for_generalization(
                    root_package, conversion, generalization)
                graph += owlutils.sub_class_of(class_uri, general_uri)
                conversion.register_element(generalization, None)
                conversion.uml_subtypes_by_uml_type.setdefault(
                    general, set()).add(class_)

    def _get_graph_for_generalization(
            self, root_package, conversion, generalization):
        """:todo Doc:"""

        return self._get_graph_for_package(
            root_package, conversion, generalization.uml_specific.uml_package)



# TODO end final

    def _convert_properties(self, root_package, conversion):
        """:todo Doc:"""

        properties, associations = (
            self._get_properties_and_associations_in_scope(
                root_package, conversion))

        for items, converter in [
                (properties, self._convert_non_assoc_property),
                (associations, self._convert_properties_of_association)]:
            items = apply_while_progress(
                items=items,
                function=lambda item, converter=converter: converter(
                    root_package, conversion, item, is_final_run=False))
            for item in items:
                handled = converter(
                    root_package, conversion, item, is_final_run=True)
                assert handled


    def _get_properties_and_associations_in_scope(
            self, root_package, conversion):
        """:todo Doc:"""

        associations = list(itertools.chain.from_iterable(
            (packaged_element
               for packaged_element in nested_package.uml_packagedElement
               if type_is(packaged_element, uml.Association))
            for nested_package in self._iter_nested_packages(root_package)))

        # This is for membership testing only, see below.
        associations_as_set = set(associations)
        assert len(associations) == len(associations_as_set)

        properties = []
        for nested_package in self._iter_nested_packages(root_package):
            for packaged_element in nested_package.uml_packagedElement:
                if isinstance(packaged_element, (uml.Class, uml.DataType)):
                    for owned_attribute in packaged_element.uml_ownedAttribute:
                        if type_is(owned_attribute, uml.Property):
                            if owned_attribute.uml_association:
                                if (owned_attribute.uml_association
                                        not in associations_as_set):
                                    WARN
                            else:
                                properties.append(owned_attribute)
                         #       print(owned_attribute.uml_name)
                                
        return properties, associations


    def _convert_non_assoc_property(
            self, root_package, conversion, property_, is_final_run):
        """:todo Doc:"""

        if property_.uml_redefinedProperty and property_.uml_subsettedProperty:
            TODO
            return True
        elif property_.uml_subsettedProperty:
            TODO
        elif property_.uml_redefinedProperty:
            
            return self._convert_non_assoc_property_redefining(
                root_package, conversion, property_, is_final_run)
        else:
            return self._convert_non_assoc_property_non_redefining(
                root_package, conversion, property_, is_final_run)


    def _convert_non_assoc_property_non_redefining(
            self, root_package, conversion, property_, is_final_run):
        """:todo Doc:"""

        type_uri, owl_property_type = (
            self._get_type_uri_and_owl_property_type_for_non_assoc_property(
                conversion, property_, {}))
        assert owl_property_type
       # if (type_uri, owl_property_type) == (None, None):
       #     return True
       # assert type_uri and owl_property_type

        # TODO: check/warn inconsistent redefinitions:
        # for each subclass of prop class (including indirect subclasses),
        # there must be exactly one unambiguous 'variant' of this property
        # (either a redefinition, or this property itself). 

        property_uri = self._create_uri_for_standalone_property(
            root_package, conversion, property_)
        conversion.register_element(property_, property_uri)

        upper = property_.uml_upper
        lower = property_.uml_lower

        if upper == lower == 0:
            ko
            conversion.add_warning(
                property_,
                'This Property (which is not a redefinition of another '
                  'Property) has an upper limit of 0. Thus, using the '
                  'corresponding OWL property will cause an inconsistent '
                  'ontology.')
        elif upper != '*' and upper < lower:
            ko
            conversion.add_warning(
                property_,
                'This Property has a lower limit of {lower} and an upper '
                    'limit of {upper}. Thus, the domain class of the '
                    'corresponding OWL property is equivalent to '
                    'owl:Nothing (i.e., there must not be any instances '
                    'of the domain class).'
                    .format(**locals()))

        class_uri = conversion.get_element_uri(property_.uml_class)

        graph = self._get_graph_for_standalone_property(
            root_package, conversion, property_)


        graph += owlutils.define_property(
            property_uri,
            owl_property_type,
            domain=class_uri,
            range=type_uri,
            min_cardinality=lower,
            max_cardinality=upper)

        return True


    def _convert_non_assoc_property_redefining(
            self, root_package, conversion, property_, is_final_run):
        """:todo Doc:"""
        
        if not is_final_run:
            return False
        
        assert property_.uml_name == 'Value'

        for redefined_property in property_.uml_redefinedProperty:
            assert redefined_property.uml_name == 'Value'
            continue
            if redefined_property not in conversion.uri_by_element:
                if is_final_run:
                    conversion.add_warning(
                        property_,
                        '''This Property has been skipped because at least one
                            the Properties that it redefines has not been
                            converted (e.g., because it is not in scope or
                            because it has been skipped).''')
                    return True
                else:
                    return False

        redefined_property_closure = closure(
            [property_],
            uml.Property.uml_redefinedProperty.get)
        redefined_property_closure.discard(property_)
        assert redefined_property_closure == set((redefined_property, ))
        root_properties = [
            redefined_prop for redefined_prop in redefined_property_closure
            if not redefined_prop.uml_redefinedProperty] # TODO test 2
        assert root_properties== [redefined_property]
        if len(root_properties) > 1:
            TODO
            
        property_uri = conversion.get_element_uri(root_properties[0])              

        class_generals = self._get_generals_in_scope(
            conversion, property_.uml_class)
        if not class_generals.issuperset(
                redefined_property.uml_class
                for redefined_property in redefined_property_closure):
            conversion.add_warning(
                property_,
                'This Property has been skipped because it redefines a '
                    'Property whose class is not a superclass of this '
                    'Property\'s class.')
            ko
            return True
        class_uri = conversion.get_element_uri(property_.uml_class)
        
        class_specializations = closure(
            [property_.uml_class],
            uml.A_superClass_class.uml_class.get)
        
        # raise Exception(class_specializations)

        # TODO: test empty name
        if property_.uml_name is None:
            WARNING
        else:
            for redefined_property in property_.uml_redefinedProperty:
                if (redefined_property.uml_name is not None
                        and property_.uml_name != redefined_property.uml_name):
                    
                    WARNINGSKIP
                    return True
    
        type_uri, owl_property_type = (
            self._get_type_uri_and_owl_property_type_for_non_assoc_property(
                conversion, property_, redefined_property_closure))
        if (type_uri, owl_property_type) == (None, None):
            return ko
        assert type_uri and owl_property_type
        # TODO: check type consistent?
        # Skip type if inherited.
        if any(property_.uml_type == redefined_property.uml_type
                for redefined_property in redefined_property_closure):
            type_uri = None

        inherited_lowers = set(
            redefined_property.uml_lower
            for redefined_property in redefined_property_closure)
        if property_.uml_lower < max(inherited_lowers):
            conversion.add_warning(
                property_,
                'The lower limit of this Property has been ignored because it '
                    'is not consistent with the lower limits of the redefined '
                    'Properties.')
            lower = None
        elif property_.uml_lower in inherited_lowers:
            lower = None
        else:
            lower = property_.uml_lower

        inherited_uppers = set(
            redefined_property.uml_upper
            for redefined_property in redefined_property_closure)
        if property_.uml_upper == '*':
            if inherited_uppers != {'*'}:
                conversion.add_warning(
                    property_,
                    'The upper limit of this Property is not consistent with '
                    'the upper limits of the redefined Properties.')
            upper = None
        else:
            inherited_uppers.discard('*')
            if inherited_uppers and property_.uml_upper > min(inherited_uppers):
                conversion.add_warning(
                    property_,
                    'The upper limit of this Property has been ignored because '
                        'it is not consistent with the upper limits of the '
                        'redefined Properties.')
                upper = None
            elif property_.uml_upper in inherited_uppers:
                upper = None
            else:
                upper = property_.uml_upper



        # TODO: refactor with non-redef version?
        # TODO: test both checks: the one below, the zero test which should not trigger here
        if (property_.uml_upper != '*'
                and property_.uml_upper < property_.uml_lower):
            conversion.add_warning(
                property_,
                'This Property has a lower limit of {property_.uml_upper} and an upper '
                    'limit of {property_.uml_upper}. Thus, the domain class of the '
                    'corresponding OWL property is equivalent to '
                    'owl:Nothing (i.e., there must not be any instances '
                    'of the domain class).'
                    .format(**locals()))


        conversion.register_element(property_, None)
        
                
            
# TODO: see also Property::isConsistentWith
            
# TODO: test prop redefines itself!
        

        # TODO: check consistency
        #       - check name
        #       - within prop (like in sibling method)
        #       - with redefined classes



        graph = self._get_graph_for_non_assoc_property_redefining(
            root_package, conversion, property_)

        graph += owlutils.restrict_class_on_property(
            class_uri,
            property_uri,
            owl_property_type,
            all_values_from=type_uri,
            min_cardinality=lower,
            max_cardinality=upper)

        return True


    def _get_type_uri_and_owl_property_type_for_non_assoc_property(
            self, conversion, property_, redefined_property_closure):
        # pylint: disable=invalid-name
        """:todo Doc:"""

        type_ = property_.uml_type
        if not type_:
            assert property_.uml_owner.uml_name == 'CustomAttribute'
            #raise Exception(property_.uml_name)
           # conversion.add_warning(
            #    property_,
            #    'This Property has been skipped because no type is given.')
            return None, OWL.ObjectProperty
        elif 0 and isinstance(type_, uml.PrimitiveType):
            ko
            type_uri = self._owl_type_by_uml_primitive_type.get(type_)
            if type_uri:
                if any(redefined_property.uml_type != type_
                        for redefined_property in redefined_property_closure):
                    conversion.add_warning(
                        property_,
                        'This Property has been skipped because its type is '
                            'not consistent with the types of its redefined '
                            'Properties.')
                    return None, None
                return type_uri, OWL.DatatypeProperty
            else:
                conversion.add_warning(
                    property_,
                    'This Property has been skipped because no OWL translation '
                        'for its type is known.')
                return None, None
        elif 1:#type_is(type_, (uml.Class, uml.Enumeration)):
            type_uri = conversion.get_element_uri(type_)
            assert type_uri
            if type_uri:
                if redefined_property_closure:
                    type_generals = self._get_generals_in_scope(
                        conversion, type_)
                    redefined_types = {
                        redefined_property.uml_type
                        for redefined_property in redefined_property_closure}
                    redefined_types.discard(None)
                    if not type_generals.issuperset(redefined_types):
                        print('---', ', '.join(t.uml_name for t in redefined_types))
                        print('---', ', '.join(t.uml_name for t in type_generals))
                        conversion.add_warning(
                            property_,
                            'This Property has been skipped because its type '
                                'is not consistent with the types of its '
                                'redefined Properties.')
                        return None, None
                return type_uri, OWL.ObjectProperty
            else:
                ko
                conversion.add_warning(
                    property_,
                    'This Property has been skipped because its type is not in '
                        'the scope of the OWL conversion (e.g., because it is '
                        'in a Package that is not converted).')
                return None, None
        else:
            raise Exception(type_)
            conversion.add_warning(
                property_,
                'This Property has been skipped because its type is not '
                'supported by this UML2OWL converter.')
            return None, None



        
        
        
        
    def _convert_properties_of_association(self, root_package, conversion, association, is_final_run):
        # TODO: ok???
        for prop in association.uml_memberEnd:
            assert prop.uml_name
            return self._convert_non_assoc_property(root_package, conversion, prop, is_final_run)
            break
        
        return True


    def _get_generals_in_scope(self, conversion, classifier):
        #assert type_is(classifier, (uml.Class, uml.Enumeration))
        return closure(
            [classifier],
            lambda classifier: [
                generalization.uml_general
                for generalization in classifier.uml_generalization
                if generalization in conversion.uri_by_element])        
        

    def _complete_non_assoc_property(self, prop):
         
        if prop.uml_subsettedProperty:
            TODO
             
        if prop.uml_redefinedProperty:
            # TODO: closure
            prop_uris = set(
                self._get_element_uri(redefined_prop, required=False)
                for redefined_prop in prop.uml_redefinedProperty)
            if None in prop_uris:
                # Wait until all redefined properties are complete.
                return False
            if len(prop_uris) != 1:
                INCONSISTENT_ERROR
            prop_uri = prop_uris.pop()
 
        # TODO: move
        graph = self._get_target_graph(prop.uml_owner.uml_package)
 
        class_ = prop.uml_owner
        assert isinstance(class_, uml.Class)
        class_uri = self._get_element_uri(class_)
 
        type_ = prop.uml_type
        # TODO: move
        if isinstance(type_, (uml.Enumeration, uml.Class)):
            type_uri = self._get_element_uri(type_)
        else:
            type_uri = {
                uml.String: XSD.string,
                uml.Integer: XSD.integer,
                uml.Real: XSD.double}.get(type_)
            if not type_uri:
                raise Exception(type_.uml_name, prop.uml_name)
 
        if not prop.uml_redefinedProperty:
            prop_uri = self._new_property_uri(prop)
            self._set_element_uri(prop, prop_uri)
            if isinstance(prop.uml_type, uml.Enumeration):
                owl_type = OWL.ObjectProperty
            else:
                owl_type = OWL.DatatypeProperty
            rdfutils.add_predicates_objects(graph, prop_uri, [
                (RDF.type, owl_type),
                (RDFS.domain, class_uri),
                (RDFS.range, type_uri)])
 
        # TODO: special case lower == upper
        restrict_all_values_from(graph, class_uri, prop_uri, type_uri)
        if prop.uml_lower > 0:
            restrict_min_cardinality(graph, class_uri, prop_uri, prop.uml_lower)
        if prop.uml_upper != '*':
            restrict_max_cardinality(graph, class_uri, prop_uri, prop.uml_upper)
             
        if isinstance(type_, (uml.Enumeration, uml.Class)):
            # TODO: other graph?
            restrict_all_values_from(graph, type_uri, prop_uri, class_uri, inverse=True)
 
 
        return True

                
                
        
        

                                    


########













        


        
        




        



    # TODO END OK

# 
#     def _complete_properties(self):
#         """:todo Doc:"""
# 
#         to_visit = set(self._properties)
#         while to_visit:
#             next_to_visit = set()
#             for prop in to_visit:
#                 if not self._complete_property(prop):
#                     next_to_visit.add(prop)
#             if len(next_to_visit) == len(to_visit):
#                 NOPROGRESSERROR
#             to_visit = next_to_visit
# 
# 
#     def _complete_property(self, prop):
#         """:todo Doc:"""
#         
#         if prop.uml_association:
#             return self._complete_assoc_property(prop)
#         else:
#             return self._complete_non_assoc_property(prop)
#         
#     def _complete_assoc_property(self, prop):
# 
#         assoc = prop.uml_association
# 
#         index = assoc.uml_memberEnd.index(prop)
#         assert index in [0, 1]
#         if index != 0:
#             return True
#         
#         # TODO: use uml api
#         opposite = assoc.uml_memberEnd[1-index]
#         
#         if prop.uml_subsettedProperty or opposite.uml_subsettedProperty:
#             TODO
#             
#         if prop.uml_redefinedProperty:
#             COV
#             if len(prop.uml_redefinedProperty) > 1:
#                 TODO
#             # TODO: closure
#             
#             redefined_prop = list(prop.uml_redefinedProperty)[0]
#             
#             if redefined_prop.uml_name:
#                 COV
#                 prop_uri = self._get_element_uri(redefined_prop, required=False)
#                 if not prop_uri:
#                     COV
#                     return False
#                 prop_uri_is_inverse = False
#                 COV
#             else:
#                 COV
#                 # TODO: check if assoc.
#                 redefined_opp = [x for x in redefined_prop.uml_owner.uml_memberEnd if x is not redefined_prop][0]
#                 assert redefined_opp.uml_name
#                 prop_uri = self._get_element_uri(redefined_opp, required=False)
#                 if not prop_uri:
#                     COV
#                     return False
#                 COV
#                 prop_uri_is_inverse = True
#                 
#         if opposite.uml_redefinedProperty:
#             TODO 
#             
#         # TODO: move
#         graph = self._get_target_graph(prop.uml_owner.uml_package)
# 
#         class_ = prop.uml_owner
#         assert isinstance(class_, uml.Class)
#         class_uri = self._get_element_uri(class_)
# 
#         type_ = prop.uml_type
#         assert isinstance(class_, uml.Class)
#         type_uri = self._get_element_uri(type_)
#         
#         
#         if not prop.uml_redefinedProperty:
#             
#             if prop.uml_name:
#             
#                 prop_uri = self._new_property_uri(prop)
#                 prop_uri_is_inverse = False
#                 self._set_element_uri(prop, prop_uri)
#                 rdfutils.add_predicates_objects(graph, prop_uri, [
#                     (RDF.type, OWL.ObjectProperty),
#                     (RDFS.domain, class_uri),
#                     (RDFS.range, type_uri)])
#                 
#                 
#                 if graph.identifier.endswith('/dexpi.owl'):
#                     DEXPI = rdflib.Namespace(graph.identifier+self._owl_sep)
#                     if prop.uml_aggregation == uml.AggregationKind.composite:
#                         graph.add((prop_uri, RDFS.subPropertyOf, DEXPI.IsComposedOf))
#                 
#             
#                 
#             else:
#                 TODO
#                 
#         if not opposite.uml_redefinedProperty:
#             
#             if opposite.uml_name:
#                 
#                 TODO
# 
#             else:
#                 opp_uri = prop_uri
#                 opp_uri_is_inverse = True
#                 
#                 
#         # TODO: special case lower == upper
#         restrict_all_values_from(graph, class_uri, prop_uri, type_uri, inverse=prop_uri_is_inverse)
#         if prop.uml_lower > 0:
#             restrict_min_cardinality(graph, class_uri, prop_uri, prop.uml_lower, inverse=prop_uri_is_inverse)
#         if prop.uml_upper != '*':
#             restrict_max_cardinality(graph, class_uri, prop_uri, prop.uml_upper, inverse=prop_uri_is_inverse)
#                  
#                     
#         # TODO: special case lower == upper
#         # TODO: other graph?
#         restrict_all_values_from(graph, type_uri, opp_uri, class_uri, inverse=opp_uri_is_inverse)
#         if opposite.uml_lower > 0:
#             restrict_min_cardinality(graph, type_uri, opp_uri, opposite.uml_lower, inverse=opp_uri_is_inverse)
#         if opposite.uml_upper != '*':
#             restrict_max_cardinality(graph, type_uri, opp_uri, opposite.uml_upper, inverse=opp_uri_is_inverse)                   
# 
# 
#         return True         
#                 
#                 
#             
# 
#         
#         
# 
# 
#         
#                 


#     
#     

# 
# 
#         
#             
#   
#         
#         
# 
#         
#           
#           #  elif isinstance(packaged_element, uml.Class):
#           #      self._on_class(packaged_element)
#                     
#     
# 
# 
# 
#     


#     def _new_property_uri(self, prop):
#         owner = prop.uml_owner
#         namespace = self._get_target_graph(self._get_target_package(owner.uml_package))
#         
#         if isinstance(owner, uml.Class):
#             name = prop.uml_name + '_of_' + owner.uml_name
#         else:
#             TODO  
#             
#         return self._new_uri(name, namespace)

#         
#     def _get_owning_package(self, element):
#         
# 
#         package = element.uml_owner
#         if not isinstance(package, uml.Package):
#             error
#             # TODO?
#             
#         return package
#     
# 
#         

  
#                 
#             
#         
#         
#         
#     def _on_begin_element(self, element):
#         if isinstance(element, uml.Class):
#             self._on_class(element)
#         elif isinstance(element, uml.Property):
#             self._on_property(element)            
#         elif isinstance(element, uml.Enumeration):
#             self._on_enumeration(element)       
# 
#     def _get_graph(self, element):
#         return self._graph
#     
#     def _get_uri(self, element):
#         return getattr(DEXPI, element.uml_name)
#     
#     
#     
#     def _on_property(self, prop):
#         
#         if prop.uml_association:
#             return
#         
#         assert not prop.uml_redefinedProperty, prop.uml_redefinedProperty
#         assert not prop.uml_subsettedProperty, prop.uml_subsettedProperty
#         
#         graph = self._get_graph(prop)
#         uri = self._get_uri(prop) # TODO mangle
#         
#         uri = getattr(DEXPI, prop.uml_name[0].lower() + prop.uml_name[1:])
#         
#         return
#         
#         
#         
#        # TODO assert uri not in self._graph.all_nodes(), uri
#         
# 
#         if isinstance(prop.uml_type, (uml.Class, uml.Enumeration)):
#             owl_type = OWL.ObjectProperty
#             owl_range = self._get_uri(prop.uml_type)
#         else:
#             owl_type = OWL.DatatypeProperty
#             owl_range = {
#                 uml.String: XSD.string,
#                 uml.Integer: XSD.integer,
#                 uml.Real: XSD.double}[prop.uml_type]
#               
#         graph.add((uri, RDF.type, owl_type))
# 
#         owner = prop.uml_owner
#         assert isinstance(owner, uml.Class)
# 
#         restr = rdflib.BNode()
#         self._graph.add((restr, RDF.type, OWL.Restriction))
#         self._graph.add((restr, OWL.onProperty, uri))
#         self._graph.add((restr, OWL.allValuesFrom, owl_range))
#         self._graph.add((self._get_uri(owner), RDFS.subClassOf, restr))
# 
#         upper = prop.uml_upper
#         if upper == '*':
#             pass
#         else:
#             restr = rdflib.BNode()
#             self._graph.add((restr, RDF.type, OWL.Restriction))
#             self._graph.add((restr, OWL.onProperty, uri))
#             self._graph.add((restr, OWL.maxCardinality, rdflib.Literal(upper, datatype=XSD.nonNegativeInteger)))
#             self._graph.add((self._get_uri(owner), RDFS.subClassOf, restr))       
#         
#         
#         
#         
# 
#             
#         
# 
# 
# 
#             
# 
#     def _get_target_graph(self, package):
#         target_package = self._get_target_package(package)
#         target_graph = self._graph_by_package.get(target_package)
#         if not target_graph:
#             uri = self._new_package_uri(target_package)
#             target_graph = rdflib.Graph(identifier=uri, store=self._graph.store)
#             target_graph.add((uri, RDF.type, OWL.Ontology))
#             self._graph_by_package[target_package] = target_graph
#             
#             
#             # TODO: move
#             
#             self._graph.bind(target_package.uml_name, uri+self._owl_sep)
#             
#             if uri.endswith('/dexpi.owl'):
#                 
#                 DEXPI = rdflib.Namespace(uri+self._owl_sep)
#                 
#                 target_graph.add((DEXPI.IsComposedOf, RDF.type, OWL.ObjectProperty))
#                 target_graph.add((DEXPI.IsComposedOf, RDF.type, OWL.InverseFunctionalProperty))
#                 target_graph.add((DEXPI.IsComponentOf, RDF.type, OWL.ObjectProperty))
#                 target_graph.add((DEXPI.IsComponentOf, RDF.type, OWL.FunctionalProperty))                
#                 target_graph.add((DEXPI.IsComposedOf, OWL.inverseOf, DEXPI.IsComponentOf))
# 
#                 
#                 
#                 
# 
#             
#         return target_graph
#             
#             
#                     
# 
# 
#     #    uri = self._new_package_uri(package)
#     #    self._set_element_uri(package, uri)
#         
#       #  package_graph = rdflib.Graph(identifier=uri, store=self._graph.store)
#       #  package_graph.add((uri, RDF.type, OWL.Ontology))
#         
#       #  self._graph_by_package[package] = package_graph
#             
#             
#             
#             
#             
#             
#         
#         
#         
#         
#         
# 
# 
#         
#         
#     def _set_element_uri(self, element, uri):
#         assert element not in self.uri_by_element
#         assert uri not in self._element_by_uri
#         self._element_by_uri[uri] = element
#         self.uri_by_element[element] = uri
#         
#         
# 
#                 
#                 

#                 
#                 
#                 
# 
#          #   super_cls_uri = self.uri_by_element[super_cls] # TODO
#          #   graph.add((cls_uri, RDFS.subClassOf, super_cls_uri))
# 
#         for prop in cls.uml_ownedAttribute:
#             
#             if prop.uml_association:
#                 continue
#             
#             assert not prop.uml_redefinedProperty
#             assert not prop.uml_subsettedProperty
#             
#             
#             
#             continue
#             
#             
#             # TODO: avoid of when possible
#             prop_uri = getattr(DEXPI, prop.uml_name + '_of_' + element.uml_name)
#             assert prop_uri not in self._graph.all_nodes()
#             
#             if prop.uml_type in (unicode, int, float):
#                 prop_type = OWL.DatatypeProperty
#  
#             else:
#                 prop_type = OWL.ObjectProperty
#                 prop_range = cls_uri(prop.uml_type)
#                 if isinstance(prop.uml_type, uml.Class):
#                     pass#TODOCLASS
#                 elif isinstance(prop.uml_type, uml.Enumeration):
#                     pass#TODOCLASS
#                 else:
#                     raise Exception(prop.uml_type)
#                 
#             
#                 
#             self._graph.add((prop_uri, RDF.type, prop_type))
#             self._graph.add((prop_uri, RDFS.domain, cls))
#             self._graph.add((prop_uri, RDFS.range, prop_range))
#             
#             upper = prop.uml_upper
#             if upper == '*':
#                 pass
#             else:
#                 assert isinstance(upper, int) and upper >= 1
#                 restr = rdflib.BNode()
#                 self._graph.add((restr, RDF.type, OWL.Restriction))
#                 self._graph.add((restr, OWL.onProperty, prop_uri))
#                 self._graph.add((restr, OWL.maxCardinality, rdflib.Literal(upper, datatype=XSD.nonNegativeInteger)))
#                 self._graph.add((cls, RDFS.subClassOf, restr))
#                 
# 
#             lower = prop.uml_lower
#             
#             if lower != 0:
#                 assert isinstance(upper, int) and lower >= 0
#                 restr = rdflib.BNode()
#                 self._graph.add((restr, RDF.type, OWL.Restriction))
#                 self._graph.add((restr, OWL.onProperty, prop_uri))
#                 self._graph.add((restr, OWL.minCardinality, rdflib.Literal(lower, datatype=XSD.nonNegativeInteger)))
#                 self._graph.add((cls, RDFS.subClassOf, restr))
#                 
#             if prop.uml_aggregation == 'composite':
#                 self._graph.add((
#                     prop_uri, RDFS.subPropertyOf, BASE.isComposedOf))
#                 
# 
# 
# 
#     
#     def _on_enumeration(self, element):
#         return
#         
#         cls = cls_uri(element)
#         
#         
# 
#         type_triple = (cls, RDF.type, OWL.Class)
#         assert type_triple not in self._graph
#         self._graph.add(type_triple)
#         
#         self._subclasses.setdefault(element, set())
#         
#         lit_uris = set()
# 
#         for lit in element.uml_ownedLiteral:
#             lit_uri = getattr(DEXPI, element.uml_name + '_' + lit.uml_name)
#             self._graph.add((lit_uri, RDF.type, cls))
#             lit_uris.add(lit_uri)
#             
#         assert lit_uris
#         
#         import rdflib.collection
#         
#         lst = rdflib.BNode()
#         
#         
#             
#         rdflib.collection.Collection(
#             self._graph, lst, lit_uris)
#         
#         self._graph.add((cls, OWL.oneOf, lst))
#         
#         all_diff = rdflib.BNode()
#         self._graph.add((all_diff, RDF.type, OWL.AllDifferent))
#         
#         lst = rdflib.BNode()
#         self._graph.add((all_diff, OWL.members, lst))
#         
#         rdflib.collection.Collection(
#             self._graph, lst, lit_uris)
#         



def uml2owl(package, **config):
    converter = Uml2OwlConverter(Uml2OwlConverterConfig(**config))
    return converter.run(package)

def create_owl(element):
    
    
    element.uml_URI = 'http://www.aixcape.org/integration'
    conversion = uml2owl(element)

 #   OwlBuilder(element)

  #  g = rdflib.Graph()
  #  g.add((DEXPI.Equipment, RDF.type, DEXPI.Murks))
  
  
    graph = conversion.owl_graph

   # for context in graph.contexts():
        
    #    print(context)
    write_owl_graphs(graph, r'C:\temp\owl')
    return graph
    


    
def write_owl_graphs(graph, root_dir):
    
    for context in graph.contexts():
        
        # ParseResult(scheme=u'http', netloc=u'www.aixcape.org', path=u'/integration/iso_15926.owl', params='', query='', fragment='')
        parse_result = urlparse.urlparse(context.identifier)
        path_parts = [root_dir, parse_result.netloc] + parse_result.path.split('/')
        path = upath.absjoin(*path_parts)
        assert path.startswith(root_dir)
        upath.makedirs(upath.dirname(path))
        context.serialize(destination=path, format='xml')
        context.serialize(destination=path+'.n3', format='n3')
        g = rdflib.Graph().parse(path+'.n3', format='n3')
        g.serialize(destination=path, format='pretty-xml')
        


        
# TODO: move

def with_indef_article(noun):
    if noun[0].lower() in 'aeiou':
        article = 'an'
    else:
        article = 'a'
    return '{article} {noun}'.format(**locals())
        
        
'''
in DEXPI
--------
            
            if uri.endswith('/dexpi.owl'):
                
                DEXPI = rdflib.Namespace(uri+self._owl_sep)
                
                target_graph.add((DEXPI.IsComposedOf, RDF.type, OWL.ObjectProperty))
                target_graph.add((DEXPI.IsComposedOf, RDF.type, OWL.InverseFunctionalProperty))
                target_graph.add((DEXPI.IsComponentOf, RDF.type, OWL.ObjectProperty))
                target_graph.add((DEXPI.IsComponentOf, RDF.type, OWL.FunctionalProperty))                
                target_graph.add((DEXPI.IsComposedOf, OWL.inverseOf, DEXPI.IsComponentOf))

'''     


def type_is(obj, type_):
    try:
        iter(type_)
        return type(obj) in type_
    except TypeError:
        return type(obj) is type_



def apply_while_progress(items, function):
    progress = True
    while items and progress:
        unhandled_items = []
        for item in items:
            item_progress = function(item)
            assert isinstance(item_progress, bool)
            if not item_progress:
                unhandled_items.append(item)
        progress = len(unhandled_items) < len(items)
        items = unhandled_items
    return items


def closure(values, getter):
    
    visited = set()
    to_visit = set(values)
    
    while to_visit:
        item = to_visit.pop()
        visited.add(item)
        to_visit.update(set(getter(item)).difference(visited))
        
    return visited








class AttributeList(list):
    pass


def a_class(node):
    return [(node, RDF.type, OWL.Class)]

def complement_of(cls, other):
    return [(cls, OWL.complementOf, other)]

def a_named_individual(node, class_=None):
    result = [(node, RDF.type, OWL.NamedIndividual)]
    if class_ is not None:
        result += [(node, RDF.type, class_)]
    return result
        

def a(node, node_type):
    return [(node, RDF.type, node_type)]

def sub_class_of(sub_class, super_class):
    return [(sub_class, RDFS.subClassOf, super_class)]



def sub_property_of(sub_class, super_class):
    return [(sub_class, RDFS.subPropertyOf, super_class)]






def equivalent_class(class_1, class_2):
    return [(class_1, OWL.equivalentClass, class_2)]

def equivalent_union(class_, sub_classes):
    union_of_result = union_of(sub_classes)
    result = (
        union_of_result +
        equivalent_class(class_, union_of_result.class_))
    return result

def equivalent_one_of(class_, individuals):
    one_of_result = one_of(individuals)
    result = (
        one_of_result +
        equivalent_class(class_, one_of_result.class_))
    return result

def all_different(individuals):
    all_different = rdflib.BNode()
    individual_list = rdflist(individuals)
    result = [(all_different, OWL.distinctMembers, individual_list.first)]
    result += a(all_different, OWL.AllDifferent)
    result += individual_list
    return result

def union_of(classes):
    class_ = rdflib.BNode()
    class_list = rdflist(classes)
    result = AttributeList(itertools.chain(
        a(class_, OWL.Class),
        [(class_, OWL.unionOf, class_list.first)],
        class_list))
    result.class_ = class_
    return result


def intersection_of(classes):
    class_ = rdflib.BNode()
    class_list = rdflist(classes)
    result = AttributeList(itertools.chain(
        a(class_, OWL.Class),
        [(class_, OWL.intersectionOf, class_list.first)],
        class_list))
    result.class_ = class_
    return result

def one_of(individuals):
    class_ = rdflib.BNode()
    individual_list = rdflist(individuals)
    result = AttributeList(itertools.chain(
        a(class_, OWL.Class),
        [(class_, OWL.oneOf, individual_list.first)],
        individual_list))
    result.class_ = class_
    return result
       
def rdflist(items):
    result = AttributeList()
    first_node = None
    prev_node = None
    for item in items:
        node = rdflib.BNode()
        if prev_node:
            result.append((prev_node, RDF.rest, node))
        else:
            first_node = node
        prev_node = node
        result.append((node, RDF.first, item))
    if first_node:
        result.first = first_node
        result.append((node, RDF.rest, RDF.nil))
    else:
        result.first = RDF.nil
    return result
        
        
        
def all_disjoint_classes(classes):
    axiom = rdflib.BNode()
    members = rdflist(classes)
    result = a(axiom, OWL.AllDisjointClasses)
    result.append((axiom, OWL.members, members.first))
    result.extend(members)
    return result
    
    result.extend()
    
    
    
    
def restrict_some_values_from(restricted_class, prop, property_type, values_class, inverse=False):
    return _restrict_property(restricted_class, prop, property_type, [(OWL.someValuesFrom, values_class)], inverse)

def restrict_all_values_from(restricted_class, prop, property_type, values_class, inverse=False):
    return _restrict_property(restricted_class, prop, property_type, [(OWL.allValuesFrom, values_class)], inverse)

def restrict_cardinality(restricted_class, prop, property_type, cardinality, inverse=False):
    return _restrict_property(restricted_class, prop, property_type, [(OWL.cardinality, rdflib.Literal(cardinality, datatype=rdflib.XSD.nonNegativeInteger))], inverse)

def restrict_min_cardinality(restricted_class, prop, property_type, min_cardinality, inverse=False):
    return _restrict_property(restricted_class, prop, property_type, [(OWL.minCardinality, rdflib.Literal(min_cardinality, datatype=rdflib.XSD.nonNegativeInteger))], inverse)
     
def restrict_max_cardinality(restricted_class, prop, property_type, max_cardinality, inverse=False):
    return _restrict_property(restricted_class, prop, property_type,
        [(OWL.maxCardinality, rdflib.Literal(max_cardinality, datatype=rdflib.XSD.nonNegativeInteger))], inverse)
    
def restrict_max_cardinality_qualified(restricted_class, prop, property_type, max_cardinality, qualified_type, inverse=False):
    return _restrict_property(restricted_class, prop, property_type,
        [(OWL.maxQualifiedCardinality, rdflib.Literal(max_cardinality, datatype=rdflib.XSD.nonNegativeInteger)),
         (OWL.onClass, qualified_type)], inverse)
    
def restrict_min_cardinality_qualified(restricted_class, prop, property_type, min_cardinality, qualified_type, inverse=False):
    return _restrict_property(restricted_class, prop, property_type,
        [(OWL.minQualifiedCardinality, rdflib.Literal(min_cardinality, datatype=rdflib.XSD.nonNegativeInteger)),
         (OWL.onClass, qualified_type)], inverse)
    
    
def restrict_cardinality_qualified(restricted_class, prop, property_type, cardinality, qualified_type, inverse=False):
    return _restrict_property(restricted_class, prop, property_type,
        [(OWL.cardinality, rdflib.Literal(cardinality, datatype=rdflib.XSD.nonNegativeInteger)),
         (OWL.onClass, qualified_type)], inverse)
    
    
    
def _restrict_property(restricted_class, prop, property_type, restriction_props, inverse):
    
    assert property_type in (OWL.DatatypeProperty, OWL.ObjectProperty)
    
    result = []

    if inverse:
        inv_prop = rdflib.BNode()
        result += a(inv_prop, OWL.DatatypeProperty)
        result.append((inv_prop, OWL.inverseOf, prop))
        prop = inv_prop

    restriction = rdflib.BNode()
    result += [
        (restriction, RDF.type, OWL.Restriction),
        (restricted_class, RDFS.subClassOf, restriction),
        (restriction, OWL.onProperty, prop)]
    for pred, obj in restriction_props:
        result.append((restriction, pred, obj))
        
    return result    
               


def restrict_class_on_property(
        class_node, property_node, property_type, all_values_from=None,
        min_cardinality=None, max_cardinality=None, max_symbols=(None, '*')):
    
    result = []
    
    if all_values_from is not None:
        result += restrict_all_values_from(class_node, property_node, property_type, all_values_from)
        
    if min_cardinality is not None:
        checks.check_non_negative_integer(min_cardinality)
    if max_cardinality in max_symbols:
        max_cardinality = None
    elif max_cardinality is not None:
        checks.check_non_negative_integer(max_cardinality)
        
    if None != min_cardinality == max_cardinality:
        result += restrict_cardinality(class_node, property_node, property_type, min_cardinality)
    else:
        if min_cardinality is not None and min_cardinality > 0:
            result += restrict_min_cardinality(class_node, property_node, property_type, min_cardinality)
        if max_cardinality is not None:
            result += restrict_max_cardinality(class_node, property_node, property_type, max_cardinality)            

    return result
        


def define_property(node, property_type, domain=None, range=None, min_cardinality=None,
        max_cardinality=None, max_symbols=(None, '*'), verbose=False):
    
    result = a(node, property_type)
    if domain is not None:
        result.append((node, RDFS.domain, domain))
    if range is not None:
        result.append((node, RDFS.range, range))
    if min_cardinality is not None:
        checks.check_non_negative_integer(min_cardinality, 'min_cardinality')
    if max_cardinality in max_symbols:
        max_cardinality = None
    else:
        checks.check_non_negative_integer(max_cardinality, 'max_cardinality')
    if min_cardinality == max_cardinality != None:
        # Case min_cardinality == max_cardinality == 1 is excluded; it will be
        # written using owl:FunctionalProperty and a min. cardinalty instead.
        result.extend(restrict_cardinality(domain, node, property_type, min_cardinality))
    elif 0 and max_cardinality == 1:
        result += a(node, OWL.FunctionalProperty)
        if min_cardinality > 0:
            result += restrict_min_cardinality(domain, node, property_type, min_cardinality)
    else:
        if domain is None:
            ERROR # using owl:Thing instead ok? check before if?     
        if min_cardinality > 0:
            result.extend(restrict_min_cardinality(domain, node, property_type, min_cardinality))
        if max_cardinality is not None:
            result.extend(restrict_max_cardinality(domain, node, property_type, max_cardinality))            
            

   # g = rdflib.Graph()
   # g += result
   # print(g.serialize())
    #kp

    return result


from unittest import TestCase
from pnb.mcl.utils import *


class Test_check_is_symbol(TestCase):

    def test_error_msg(self):
        with self.assertRaises(ValueError) as cm:
            check_is_symbol('xy z')
        self.assertEqual(
            str(cm.exception),
            r"'xy z' does not match the pattern [A-Za-z_][A-Za-z_0-9]*")

    def test_error_msg_name(self):
        with self.assertRaises(ValueError) as cm:
            check_is_symbol('xy z', 'test')
        self.assertEqual(
            str(cm.exception),
            r"test: 'xy z' does not match the pattern [A-Za-z_][A-Za-z_0-9]*")

    def test_ok(self):
        for text in ["A", "A1", "A_", "_", "_9", "_B", "__", "_F7"]:
            check_is_symbol(text)

    def test_error(self):
        for text in ["", " ", "7", "3a", "X ", " x", "ö", "a.a"]:
            self.assertRaises(ValueError, lambda text=text: check_is_symbol(text))


class Test_check_is_symbol_reference(TestCase):

    def test_error_msg(self):
        with self.assertRaises(ValueError) as cm:
            check_is_symbol_reference('x.y z')
        self.assertEqual(
            str(cm.exception),
            r"'x.y z' does not match the pattern "
                r"[A-Za-z_][A-Za-z_0-9]*(\.[A-Za-z_][A-Za-z_0-9]*)?")

    def test_error_msg_name(self):
        with self.assertRaises(ValueError) as cm:
            check_is_symbol_reference('x.y z', 'test')
        self.assertEqual(
            str(cm.exception),
            r"test: 'x.y z' does not match the pattern "
                r"[A-Za-z_][A-Za-z_0-9]*(\.[A-Za-z_][A-Za-z_0-9]*)?")

    def test_ok(self):
        for text in ["A", "A1", "A_", "_", "_9", "_B", "__", "_F7",
                     "A._", "A1.Q", "A_.z", "_._", "_9.k", "_B.kyl1",
                     "__.B3s", "_F7.L",
                     "dexpi.PositiveDisplacementPump"]:
            check_is_symbol_reference(text)

    def test_error(self):
        for text in ["", " ", "7", "3a", "X ", " x", "ö", "a.2",
                     "dexpi.Negative.DisplacementPump"]:
            self.assertRaises(
                ValueError,
                lambda text=text: check_is_symbol_reference(text))


class Test_check_is_multiple_symbol_references(TestCase):

    def test_error_msg(self):
        with self.assertRaises(ValueError) as cm:
            check_is_multiple_symbol_references(' x.yz')
        self.assertEqual(
            str(cm.exception),
            r"' x.yz' does not match the pattern "
                r"([A-Za-z_][A-Za-z_0-9]*(\.[A-Za-z_][A-Za-z_0-9]*)?"
                r"( [A-Za-z_][A-Za-z_0-9]*(\.[A-Za-z_][A-Za-z_0-9]*)?)*)?")

    def test_error_msg_name(self):
        with self.assertRaises(ValueError) as cm:
            check_is_multiple_symbol_references(' x.yz', 'test')
        self.assertEqual(
            str(cm.exception),
            r"test: ' x.yz' does not match the pattern "
                r"([A-Za-z_][A-Za-z_0-9]*(\.[A-Za-z_][A-Za-z_0-9]*)?"
                r"( [A-Za-z_][A-Za-z_0-9]*(\.[A-Za-z_][A-Za-z_0-9]*)?)*)?")

    def test_ok(self):
        for text in ["", "A B C", "A1 A_", "A_ _", "_ _8", "_9 A XC", "_B",
                     "__", "_F7", "A._", "A1.Q", "A_.z", "_._", "_9.k",
                     "_B.kyl1", "__.B3s", "_F7.L",
                     "mcl is amazing"]:
            check_is_multiple_symbol_references(text)

    def test_error(self):
        for text in [" "]:
            self.assertRaises(
                ValueError,
                lambda text=text: check_is_multiple_symbol_references(text))



from pnb.mcl.test.model_api import AbstractTestCase

class Test_CompositionProperty(AbstractTestCase):

    def test_init(self):
        city = self.ConcreteClass('City')
        cities = self.CompositionProperty('cities', city, 0, 5, False)
        self.assertEqual(cities.name, 'cities')
        self.assertEqual(cities.lower, 0)
        self.assertEqual(cities.upper, 5)
        self.assertIs(cities.type, city)
        self.assertFalse(cities.isOrdered)
        self.assertTrue(cities.isUnique)
        self.assertEqual(cities.oppositeLower, 0)
        self.assertEqual(cities.oppositeUpper, 1)


class Test_ReferenceProperty(AbstractTestCase):

    def test_init(self):
        city = self.ConcreteClass('City')
        cities = self.ReferenceProperty(
            'cities', city, 0, 5, False, True, 3, 7)
        self.assertEqual(cities.name, 'cities')
        self.assertEqual(cities.lower, 0)
        self.assertEqual(cities.upper, 5)
        self.assertIs(cities.type, city)
        self.assertFalse(cities.isOrdered)
        self.assertTrue(cities.isUnique)
        self.assertEqual(cities.oppositeLower, 3)
        self.assertEqual(cities.oppositeUpper, 7)

    def test_owner(self):
        city = self.ConcreteClass('City')
        cities = self.ReferenceProperty(
            'cities', 0, 5, city, False, True, 3, 7)
        self.assertIsNone(cities.owner)

    def test_owner_after_add(self):
        city = self.ConcreteClass('City')
        cities = self.ReferenceProperty(
            'cities', 0, 5, city, False, True, 3, 7)
        city.ownedAttributes.add(cities)
        self.assertIs(cities.owner, city)

    def _test_upper_greater_equal_lower(self):
        # TODO
        city = self.ConcreteClass('City')
        self.ReferenceProperty('cities', 5, 5, city, False, True, 3, 7)
        self.ReferenceProperty('cities', 10, None, city, False, True, 3, 7)
        self.assertRaisesRegex(
            ValueError,
            'mimi',
            lambda:
                self.ReferenceProperty('cities', 5, 4, city, False, True, 3, 7))

        
        
        
        





from warnings import warn
import weakref

from pnb.mcl.test.model_api import AbstractTestCase


class Test_Members(AbstractTestCase):
    """
    Tests for direct instances of Members, which are always derived.
    
    Some tests may seem trivial or redundant. They are included to cover
    different cases w.r.t. the internal update mechanism of a model
    implementation. 
    """

    def test_private_init(self):
        self.assertRaisesRegex(
            TypeError,
            "A Members object cannot be created from user code.",
            lambda: self.Members(self.Package(name='spam'), self.Namespace.members))

    def test_identity(self):
        p = self.Package('p')
        self.assertIs(
            p.ownedMembers,
            p.ownedMembers)

    def test_gc(self):
        # Works in Python 3.12. May fail in other implementations with different
        # behavior of garbage collector,
        p = self.Package('p')
        owned_members_ref = weakref.ref(p.ownedMembers)
        self.assertIsNone(owned_members_ref())

    def test_info(self):
        self.assertEqual(
            self.Package('p').ownedMembers.info,
            "Namespace.ownedMembers of Package 'p'")

    def test_repr(self):
        self.assertEqual(
            repr(self.Package('p').ownedMembers),
            "<Namespace.ownedMembers of Package 'p'>")

    def test_contains_empty(self):
        p = self.Package('p')
        self.assertNotIn('c', p.ownedMembers)
        self.assertNotIn(self.ConcreteClass('c'), p.ownedMembers)

    def test_add_contains(self):
        p = self.Package('p')
        c = p.packagedElements.add(self.ConcreteClass('c'))
        self.assertIn('c', p.ownedMembers)
        self.assertIn(c, p.ownedMembers)
        self.assertNotIn('d', p.ownedMembers)
        self.assertNotIn(self.ConcreteClass('c'), p.ownedMembers)

    def test_add_add_contains(self):
        p = self.Package('p')
        p.packagedElements.add(self.ConcreteClass('c1'))
        p.packagedElements.add(self.ConcreteClass('c2'))
        self.assertIn('c1', p.ownedMembers)
        self.assertIn('c2', p.ownedMembers)
        self.assertNotIn('c', p.ownedMembers)
        self.assertNotIn(self.ConcreteClass('c'), p.ownedMembers)

    def test_add_contains_add_contains(self):
        p = self.Package('p')
        c2 = p.packagedElements.add(self.ConcreteClass('c2'))
        self.assertIn('c2', p.ownedMembers)
        self.assertIn(c2, p.ownedMembers)
        c1 = p.packagedElements.add(self.ConcreteClass('c1'))
        self.assertIn('c1', p.ownedMembers)
        self.assertIn(c1, p.ownedMembers)
        self.assertIn('c2', p.ownedMembers)
        self.assertIn(c2, p.ownedMembers)

    def test_getattr_empty(self):
        p = self.Package('p')
        self.assertRaisesRegex(
            AttributeError,
            "Namespace.ownedMembers of Package 'p' has no attribute 'c'.",
            lambda: p.ownedMembers.c)

    def test_add_getattr(self):
        p = self.Package('p')
        c = p.packagedElements.add(self.ConcreteClass('c'))
        self.assertRaisesRegex(
            AttributeError,
            "Namespace.ownedMembers of Package 'p' has no attribute 'd'.",
            lambda: p.ownedMembers.d)
        self.assertIs(
            p.ownedMembers.c,
            c)

    def test_add_add_getattr(self):
        p = self.Package('p')
        c1 = p.packagedElements.add(self.ConcreteClass('c1'))
        c2 = p.packagedElements.add(self.ConcreteClass('c2'))
        self.assertRaisesRegex(
            AttributeError,
            "Namespace.ownedMembers of Package 'p' has no attribute 'd'.",
            lambda: p.ownedMembers.d)
        self.assertIs(
            p.ownedMembers.c1,
            c1)
        self.assertIs(
            p.ownedMembers.c2,
            c2)

    def test_add_getattr_add_getattr(self):
        p = self.Package('p')
        c2 = p.packagedElements.add(self.ConcreteClass('c2'))
        self.assertIs(
            p.ownedMembers.c2,
            c2)
        c1 = p.packagedElements.add(self.ConcreteClass('c1'))
        self.assertIs(
            p.ownedMembers.c1,
            c1)
        self.assertIs(
            p.ownedMembers.c2,
            c2)

    def test_iter_empty(self):
        p = self.Package('p')
        self.assertEqual(
            list(p.ownedMembers),
            [])

    def test_add_iter(self):
        p = self.Package('p')
        c = p.packagedElements.add(self.ConcreteClass('c'))
        self.assertEqual(
            list(p.ownedMembers),
            [c])

    def test_add_add_iter(self):
        p = self.Package('p')
        c2 = p.packagedElements.add(self.ConcreteClass('c2'))
        c1 = p.packagedElements.add(self.ConcreteClass('c1'))
        self.assertEqual(
            list(p.ownedMembers),
            [c1, c2])

    def test_add_iter_add_iter(self):
        p = self.Package('p')
        c2 = p.packagedElements.add(self.ConcreteClass('c2'))
        self.assertEqual(
            list(p.ownedMembers),
            [c2])
        c1 = p.packagedElements.add(self.ConcreteClass('c1'))
        self.assertEqual(
            list(p.ownedMembers),
            [c1, c2])

    def test_error_changed_on_iter(self):
        p = self.Package('p')
        p.packagedElements.add(self.ConcreteClass('c1'))
        with self.assertRaisesRegex(
                RuntimeError,
                'dictionary changed size during iteration'):
            for _ in p.ownedMembers:
                p.packagedElements.add(self.ConcreteClass('c2'))

    def test_len_empty(self):
        p = self.Package('p')
        self.assertEqual(
            len(p.ownedMembers),
            0)

    def test_add_len(self):
        p = self.Package('p')
        p.packagedElements.add(self.ConcreteClass('c'))
        self.assertEqual(
            len(p.ownedMembers),
            1)

    def test_add_add_len(self):
        p = self.Package('p')
        p.packagedElements.add(self.ConcreteClass('c2'))
        p.packagedElements.add(self.ConcreteClass('c1'))
        self.assertEqual(
            len(p.ownedMembers),
            2)

    def test_add_len_add_len(self):
        p = self.Package('p')
        p.packagedElements.add(self.ConcreteClass('c2'))
        self.assertEqual(
            len(p.ownedMembers),
            1)
        p.packagedElements.add(self.ConcreteClass('c1'))
        self.assertEqual(
            len(p.ownedMembers),
            2)

    def test_at_empty(self):
        p = self.Package('p')
        self.assertRaisesRegex(
            KeyError,
            "Namespace.ownedMembers of Package 'p' has no member named 'c'.",
            lambda: p.ownedMembers.at('c'))

    def test_add_at(self):
        p = self.Package('p')
        c = p.packagedElements.add(self.ConcreteClass('c'))
        self.assertRaisesRegex(
            KeyError,
            "Namespace.ownedMembers of Package 'p' has no member named 'd'.",
            lambda: p.ownedMembers.at('d'))
        self.assertIs(
            p.ownedMembers.at('c'),
            c)

    def test_add_add_at(self):
        p = self.Package('p')
        c1 = p.packagedElements.add(self.ConcreteClass('c1'))
        c2 = p.packagedElements.add(self.ConcreteClass('c2'))
        self.assertRaisesRegex(
            KeyError,
            "Namespace.ownedMembers of Package 'p' has no member named 'd'.",
            lambda: p.ownedMembers.at('d'))
        self.assertIs(
            p.ownedMembers.at('c1'),
            c1)
        self.assertIs(
            p.ownedMembers.at('c2'),
            c2)

    def test_add_at_add_at(self):
        p = self.Package('p')
        c2 = p.packagedElements.add(self.ConcreteClass('c2'))
        self.assertIs(
            p.ownedMembers.at('c2'),
            c2)
        c1 = p.packagedElements.add(self.ConcreteClass('c1'))
        self.assertIs(
            p.ownedMembers.at('c1'),
            c1)
        self.assertIs(
            p.ownedMembers.at('c2'),
            c2)

    def test_get_empty(self):
        p = self.Package('p')
        self.assertIsNone(p.ownedMembers.get('c'))
        self.assertIs(
            p.ownedMembers.get('c', self),
            self)

    def test_add_get(self):
        p = self.Package('p')
        c = p.packagedElements.add(self.ConcreteClass('c'))
        self.assertIsNone(p.ownedMembers.get('d'))
        self.assertIs(
            p.ownedMembers.get('c'),
            c)

    def test_add_add_get(self):
        p = self.Package('p')
        c1 = p.packagedElements.add(self.ConcreteClass('c1'))
        c2 = p.packagedElements.add(self.ConcreteClass('c2'))
        self.assertIsNone(p.ownedMembers.get('d'))
        self.assertIs(
            p.ownedMembers.get('c1'),
            c1)
        self.assertIs(
            p.ownedMembers.get('c2'),
            c2)

    def test_add_get_add_get(self):
        p = self.Package('p')
        c2 = p.packagedElements.add(self.ConcreteClass('c2'))
        self.assertIs(
            p.ownedMembers.get('c2'),
            c2)
        c1 = p.packagedElements.add(self.ConcreteClass('c1'))
        self.assertIs(
            p.ownedMembers.get('c1'),
            c1)
        self.assertIs(
            p.ownedMembers.get('c2'),
            c2)

    def test_names_empty(self):
        p = self.Package('p')
        self.assertEqual(
            list(p.ownedMembers.names),
            [])

    def test_add_names(self):
        p = self.Package('p')
        p.packagedElements.add(self.ConcreteClass('c'))
        self.assertEqual(
            list(p.ownedMembers.names),
            ['c'])

    def test_add_add_names(self):
        p = self.Package('p')
        p.packagedElements.add(self.ConcreteClass('c2'))
        p.packagedElements.add(self.ConcreteClass('c1'))
        self.assertEqual(
            list(p.ownedMembers.names),
            ['c1', 'c2'])

    def test_add_names_add_names(self):
        p = self.Package('p')
        p.packagedElements.add(self.ConcreteClass('c2'))
        self.assertEqual(
            list(p.ownedMembers.names),
            ['c2'])
        p.packagedElements.add(self.ConcreteClass('c1'))
        self.assertEqual(
            list(p.ownedMembers.names),
            ['c1', 'c2'])

    def test_error_changed_on_names(self):
        p = self.Package('p')
        p.packagedElements.add(self.ConcreteClass('c1'))
        with self.assertRaisesRegex(
                RuntimeError,
                'dictionary changed size during iteration'):
            for _ in p.ownedMembers.names:
                p.packagedElements.add(self.ConcreteClass('c2'))


class Test_MutableMembers(AbstractTestCase):
    """
    Tests for direct instances of MutableMembers, which are never derived.
    
    Some tests may seem trivial or redundant. They are included to cover
    different cases w.r.t. the internal update mechanism of a model
    implementation. 
    """

    def test_private_init(self):
        self.assertRaisesRegex(
            TypeError,
            "A MutableMembers object cannot be created from user code.",
            lambda: self.MutableMembers(self.Package(name='spam'), self.Namespace.members))

    def test_identity(self):
        p = self.Package('p')
        self.assertIs(
            p.packagedElements,
            p.packagedElements)

    def test_gc(self):
        # Works in Python 3.12. May fail in other implementations with other
        # different behavior of garbage collector,
        p = self.Package('p')
        owned_members_ref = weakref.ref(p.ownedMembers)
        self.assertIsNone(owned_members_ref())

    def test_info(self):
        self.assertEqual(
            self.Package('p').packagedElements.info,
            "Package.packagedElements of Package 'p'")

    def test_repr(self):
        self.assertEqual(
            repr(self.Package('p').packagedElements),
            "<Package.packagedElements of Package 'p'>")

    def test_add(self):
        p = self.Package('p')
        c = self.ConcreteClass('c')
        self.assertIs(
            p.packagedElements.add(c),
            c)
        self.assertEqual(
            list(p.packagedElements),
            [c])
        self.assertIs(
            c.owner,
            p)

    def test_add_twice(self):
        p = self.Package('p')
        c = self.ConcreteClass('c')
        self.assertIs(
            p.packagedElements.add(c),
            c)
        self.assertIs(
            p.packagedElements.add(c),
            c)
        self.assertEqual(
            list(p.packagedElements),
            [c])

    def test_error_add_wrong_type(self):
        p = self.Package('p')
        c = self.ConcreteClass('c')
        cp = self.CompositionProperty('cp', 1, 1, c, True)
        with self.assertRaises(TypeError) as cm:
            p.packagedElements.add(cp)
        self.assertEqual(
            str(cm.exception),
            "<CompositionProperty 'cp'> cannot be added to <Package.packagedElements of Package "
                "'p'> because it is not a <PackageableElement>.")

    def test_error_add_name_in_same_property(self):
        p = self.Package('p')
        p.packagedElements.add(self.ConcreteClass('c'))
        self.assertRaisesRegex(
            TypeError,
            "<ConcreteClass 'c'> cannot be added to <Package.packagedElements "
            "of Package 'p'> because <Package.packagedElements of Package 'p'> "
            "already contains a member with this name.",
            lambda: p.packagedElements.add(self.ConcreteClass('c')))

    def _test_error_add_name_in_members(self):
        TODO

    def test_error_add_other_owner(self):
        p1 = self.Package('p1')
        c = p1.packagedElements.add(self.ConcreteClass('c'))
        p2 = self.Package('p2')
        self.assertRaisesRegex(
            TypeError,
            "<ConcreteClass 'p1.c'> is already owned by <Package 'p1'> and cannot "
                "be added to <Package.packagedElements of Package 'p2'>.",
            lambda: p2.packagedElements.add(c))

    def test_error_add_other_property(self):
        warn("TODO")

    def test_contains_empty(self):
        p = self.Package('p')
        self.assertNotIn('c', p.packagedElements)
        self.assertNotIn(self.ConcreteClass('c'), p.packagedElements)

    def test_add_contains(self):
        p = self.Package('p')
        c = p.packagedElements.add(self.ConcreteClass('c'))
        self.assertIn('c', p.packagedElements)
        self.assertIn(c, p.packagedElements)
        self.assertNotIn('d', p.packagedElements)
        self.assertNotIn(self.ConcreteClass('c'), p.packagedElements)

    def test_add_add_contains(self):
        p = self.Package('p')
        p.packagedElements.add(self.ConcreteClass('c1'))
        p.packagedElements.add(self.ConcreteClass('c2'))
        self.assertIn('c1', p.packagedElements)
        self.assertIn('c2', p.packagedElements)
        self.assertNotIn('c', p.packagedElements)
        self.assertNotIn(self.ConcreteClass('c'), p.packagedElements)

    def test_add_contains_add_contains(self):
        p = self.Package('p')
        c2 = p.packagedElements.add(self.ConcreteClass('c2'))
        self.assertIn('c2', p.packagedElements)
        self.assertIn(c2, p.packagedElements)
        c1 = p.packagedElements.add(self.ConcreteClass('c1'))
        self.assertIn('c1', p.packagedElements)
        self.assertIn(c1, p.packagedElements)
        self.assertIn('c2', p.packagedElements)
        self.assertIn(c2, p.packagedElements)

    def test_getattr_empty(self):
        p = self.Package('p')
        self.assertRaisesRegex(
            AttributeError,
            "Package.packagedElements of Package 'p' has no attribute 'c'.",
            lambda: p.packagedElements.c)

    def test_add_getattr(self):
        p = self.Package('p')
        c = p.packagedElements.add(self.ConcreteClass('c'))
        self.assertRaisesRegex(
            AttributeError,
            "Package.packagedElements of Package 'p' has no attribute 'd'.",
            lambda: p.packagedElements.d)
        self.assertIs(
            p.packagedElements.c,
            c)

    def test_add_add_getattr(self):
        p = self.Package('p')
        c1 = p.packagedElements.add(self.ConcreteClass('c1'))
        c2 = p.packagedElements.add(self.ConcreteClass('c2'))
        self.assertRaisesRegex(
            AttributeError,
            "Package.packagedElements of Package 'p' has no attribute 'd'.",
            lambda: p.packagedElements.d)
        self.assertIs(
            p.packagedElements.c1,
            c1)
        self.assertIs(
            p.packagedElements.c2,
            c2)

    def test_add_getattr_add_getattr(self):
        p = self.Package('p')
        c2 = p.packagedElements.add(self.ConcreteClass('c2'))
        self.assertIs(
            p.packagedElements.c2,
            c2)
        c1 = p.packagedElements.add(self.ConcreteClass('c1'))
        self.assertIs(
            p.packagedElements.c1,
            c1)
        self.assertIs(
            p.packagedElements.c2,
            c2)

    def test_iter_empty(self):
        p = self.Package('p')
        self.assertEqual(
            list(p.packagedElements),
            [])

    def test_add_iter(self):
        p = self.Package('p')
        c = p.packagedElements.add(self.ConcreteClass('c'))
        self.assertEqual(
            list(p.packagedElements),
            [c])

    def test_add_add_iter(self):
        p = self.Package('p')
        c2 = p.packagedElements.add(self.ConcreteClass('c2'))
        c1 = p.packagedElements.add(self.ConcreteClass('c1'))
        self.assertEqual(
            list(p.packagedElements),
            [c1, c2])

    def test_add_iter_add_iter(self):
        p = self.Package('p')
        c2 = p.packagedElements.add(self.ConcreteClass('c2'))
        self.assertEqual(
            list(p.packagedElements),
            [c2])
        c1 = p.packagedElements.add(self.ConcreteClass('c1'))
        self.assertEqual(
            list(p.packagedElements),
            [c1, c2])

    def test_error_changed_on_iter(self):
        p = self.Package('p')
        p.packagedElements.add(self.ConcreteClass('c1'))
        with self.assertRaisesRegex(
                RuntimeError,
                'dictionary changed size during iteration'):
            for _ in p.packagedElements:
                p.packagedElements.add(self.ConcreteClass('c2'))

    def test_len_empty(self):
        p = self.Package('p')
        self.assertEqual(
            len(p.packagedElements),
            0)

    def test_add_len(self):
        p = self.Package('p')
        p.packagedElements.add(self.ConcreteClass('c'))
        self.assertEqual(
            len(p.packagedElements),
            1)

    def test_add_add_len(self):
        p = self.Package('p')
        p.packagedElements.add(self.ConcreteClass('c2'))
        p.packagedElements.add(self.ConcreteClass('c1'))
        self.assertEqual(
            len(p.packagedElements),
            2)

    def test_add_len_add_len(self):
        p = self.Package('p')
        p.packagedElements.add(self.ConcreteClass('c2'))
        self.assertEqual(
            len(p.packagedElements),
            1)
        p.packagedElements.add(self.ConcreteClass('c1'))
        self.assertEqual(
            len(p.packagedElements),
            2)

    def test_at_empty(self):
        p = self.Package('p')
        self.assertRaisesRegex(
            KeyError,
            "Package.packagedElements of Package 'p' has no member named 'c'.",
            lambda: p.packagedElements.at('c'))

    def test_add_at(self):
        p = self.Package('p')
        c = p.packagedElements.add(self.ConcreteClass('c'))
        self.assertRaisesRegex(
            KeyError,
            "Package.packagedElements of Package 'p' has no member named 'd'.",
            lambda: p.packagedElements.at('d'))
        self.assertIs(
            p.packagedElements.at('c'),
            c)

    def test_add_add_at(self):
        p = self.Package('p')
        c1 = p.packagedElements.add(self.ConcreteClass('c1'))
        c2 = p.packagedElements.add(self.ConcreteClass('c2'))
        self.assertRaisesRegex(
            KeyError,
            "Package.packagedElements of Package 'p' has no member named 'd'.",
            lambda: p.packagedElements.at('d'))
        self.assertIs(
            p.packagedElements.at('c1'),
            c1)
        self.assertIs(
            p.packagedElements.at('c2'),
            c2)

    def test_add_at_add_at(self):
        p = self.Package('p')
        c2 = p.packagedElements.add(self.ConcreteClass('c2'))
        self.assertIs(
            p.packagedElements.at('c2'),
            c2)
        c1 = p.packagedElements.add(self.ConcreteClass('c1'))
        self.assertIs(
            p.packagedElements.at('c1'),
            c1)
        self.assertIs(
            p.packagedElements.at('c2'),
            c2)

    def test_get_empty(self):
        p = self.Package('p')
        self.assertIsNone(p.packagedElements.get('c'))
        self.assertIs(
            p.packagedElements.get('c', self),
            self)

    def test_add_get(self):
        p = self.Package('p')
        c = p.packagedElements.add(self.ConcreteClass('c'))
        self.assertIsNone(p.packagedElements.get('d'))
        self.assertIs(
            p.packagedElements.get('c'),
            c)

    def test_add_add_get(self):
        p = self.Package('p')
        c1 = p.packagedElements.add(self.ConcreteClass('c1'))
        c2 = p.packagedElements.add(self.ConcreteClass('c2'))
        self.assertIsNone(p.packagedElements.get('d'))
        self.assertIs(
            p.packagedElements.get('c1'),
            c1)
        self.assertIs(
            p.packagedElements.get('c2'),
            c2)

    def test_add_get_add_get(self):
        p = self.Package('p')
        c2 = p.packagedElements.add(self.ConcreteClass('c2'))
        self.assertIs(
            p.packagedElements.get('c2'),
            c2)
        c1 = p.packagedElements.add(self.ConcreteClass('c1'))
        self.assertIs(
            p.packagedElements.get('c1'),
            c1)
        self.assertIs(
            p.packagedElements.get('c2'),
            c2)

    def test_names_empty(self):
        p = self.Package('p')
        self.assertEqual(
            list(p.packagedElements.names),
            [])

    def test_add_names(self):
        p = self.Package('p')
        p.packagedElements.add(self.ConcreteClass('c'))
        self.assertEqual(
            list(p.packagedElements.names),
            ['c'])

    def test_add_add_names(self):
        p = self.Package('p')
        p.packagedElements.add(self.ConcreteClass('c2'))
        p.packagedElements.add(self.ConcreteClass('c1'))
        self.assertEqual(
            list(p.packagedElements.names),
            ['c1', 'c2'])

    def test_add_names_add_names(self):
        p = self.Package('p')
        p.packagedElements.add(self.ConcreteClass('c2'))
        self.assertEqual(
            list(p.packagedElements.names),
            ['c2'])
        p.packagedElements.add(self.ConcreteClass('c1'))
        self.assertEqual(
            list(p.packagedElements.names),
            ['c1', 'c2'])

    def test_error_changed_on_names(self):
        p = self.Package('p')
        p.packagedElements.add(self.ConcreteClass('c1'))
        with self.assertRaisesRegex(
                RuntimeError,
                'dictionary changed size during iteration'):
            for _ in p.packagedElements.names:
                p.packagedElements.add(self.ConcreteClass('c2'))



def make_model_by_uri(meta_model):

    ModelCore = meta_model.Model(
        name='ModelCore',
        uri='http://ModelCore',
        members=[meta_model.AbstractClass(name='ConceptualModel')])

    ModelProcess = meta_model.Model(
        name='ModelProcess',
        uri='http://ModelProcess')
    ModelProcess.add(meta_model.Package(
        name='Base',
        members=[
            meta_model.ConcreteClass(
                name='ProcessStep',
                superTypes=[ModelCore.ConceptualModel],
                members=[
                    meta_model.DataProperty(
                        name='StepNumber',
                        lower=0,
                        upper=1,
                        type_=meta_model.MCL.String,
                        isOrdered=False,
                        isUnique=False)])]))
    ModelProcess.Base.ProcessStep.add(
        meta_model.ReferenceProperty(
            name='RelatedSteps',
            lower=0,
            upper=None,
            type_=ModelProcess.Base.ProcessStep,
            isOrdered=False,
            isUnique=True,
            oppositeLower=0,
            oppositeUpper=None))
    ModelProcess.Base.ProcessStep.add(
        meta_model.CompositionProperty(
            name='SubSteps',
            lower=0,
            upper=None,
            type_=ModelProcess.Base.ProcessStep,
            isOrdered=False))
    ModelProcess.add(meta_model.Package(
        name='Steps',
        members=[
            meta_model.ConcreteClass(
                name='Pumping',
                superTypes=[ModelProcess.Base.ProcessStep])]))

    ModelProcess.Steps.Pumping(
                StepNumber='step1')

    ModelInstance1 = meta_model.Model(
        name='ModelInstance1',
        uri='http://ModelInstance1',
        members=[
            step1:=ModelProcess.Steps.Pumping(
                StepNumber='step1')])

    ModelInstance2a = meta_model.Model(
        name='ModelInstance2',
        uri='http://ModelInstance2a',
        members=[
            step2a1:=ModelProcess.Steps.Pumping(
                name='Step1',
                StepNumber='step2a1'),
            step2a2:=ModelProcess.Steps.Pumping(
                StepNumber='step2a2')])

    ModelInstance2b = meta_model.Model(
        name='ModelInstance2',
        uri='http://ModelInstance2b',
        members=[
            step2b:=ModelProcess.Steps.Pumping(
                name='Step',
                StepNumber='step2b')])

    ModelInstance3 = meta_model.Model(
        name='ModelInstance3',
        uri='http://ModelInstance3',
        members=[
            ModelProcess.Steps.Pumping(
                StepNumber='step31',
                SubSteps=[
                    step32:=ModelProcess.Steps.Pumping(
                        name='Step',
                        StepNumber='step32'),
                    #step2a2 # TODO
                    ])])

    step1.RelatedSteps = [step2a1, step2b]
    step2a1.RelatedSteps = [step2a2, step32]
    step2b.RelatedSteps = [step2a1]

    return {model.uri: model for model in [
        ModelCore, ModelProcess, ModelInstance1, ModelInstance2a, ModelInstance2b, ModelInstance3]}
